;I2C Master library for PIC microcontrollers
;Written by Franois Gerin, 2003
;<fgerin@swing.be>
;
;This file contains assembly code for I2C communications
;using 3 pins of PIC C. (3 pins to replace "Open Drain")
;Using this code, the C becomes the MASTER device on the I2C bus.
;Multi-master mode is NOT supported.
;
;
;BE CAREFUL: You must set SCL and SDAo as output and 
;SDAi as input yourself!!! (this code doesn't verify)
;
;You need to define the following variables and functions
;before the "#include I2C_M_3pins.asm" line:
;
;(R) I2C_TEMP equ xxx   (xxx = a free register)
;(C) I2C_SDAo equ p,z   (p = port / z = pin for serial data out)
;(C) I2C_SDAi equ p,z   (p = port / z = pin for serial data in)
;(C) I2C_SCL equ p,z    (p = port / z = pin for serial clock)
;(F) I2C_DELAY      function setting the pulse width
;
;(R)=register (C)=constant (F)=function
;
;You must build the "I2C_DELAY" function yourself to
; make the pulse width according to the speed you need.
;(don't forget the four clock cycles of call and return!)
;
;
;---------------------------------
;Functions defined:
;
;I2C_START = make a Start bit
;I2C_STOP  = make a Stop bit
;I2C_M_ACK = make a Master Acknowledge bit
;I2C_M_NACK = make a Master Non-Acknowledge bit
;I2C_DEV_ACK = make a Device Acknowledge
;   (then STATUS,Z = 1 if device acknowledged
;    or STATUS,Z = 0 if device didn't.)
;
;I2C_SEND = send the byte placed in W register
;I2C_RECEIVE = get a byte and store it to W register
;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$



;COMPILER INSTRUCTIONS:
#ifndef I2C_SDAi
  ERROR I2C_SDAi not defined, see I2C_M_3pins.asm documentation!
#endif
#ifndef I2C_SDAo
  ERROR I2C_SDAo not defined, see I2C_M_3pins.asm documentation!
#endif
#ifndef I2C_SCL
  ERROR I2C_SCL not defined, see I2C_M_3pins.asm documentation!
#endif
#ifndef I2C_TEMP
  ERROR I2C_TEMP not defined, see I2C_M_3pins.asm documentation!
#endif

#define I2C_MASTER_fglib



I2C_START:

    bsf I2C_SDAo        ;I2C start: change of SDA
    call I2C_DELAY      ;from 1 to 0 while SCL = 1
    bsf I2C_SCL
    call I2C_DELAY
    bcf I2C_SDAo
    call I2C_DELAY
    bcf I2C_SCL
    call I2C_DELAY
    return
;---------------------
I2C_STOP:

    bcf I2C_SDAo        ;I2C stop: change of SDA
    call I2C_DELAY      ;from 0 to 1 while SCL = 1
    bsf I2C_SCL
    call I2C_DELAY
    bsf I2C_SDAo
    call I2C_DELAY
    return
;---------------------
I2C_M_ACK:

    bcf I2C_SDAo        ;I2C master acknowledge:
    call I2C_DELAY      ;SDA = 0 
    bsf I2C_SCL
    call I2C_DELAY
    bcf I2C_SCL
    call I2C_DELAY
    return
;---------------------
I2C_M_NACK:

    bsf I2C_SDAo        ;I2C master non-acknowledge:
    call I2C_DELAY      ;SDA = 1
    bsf I2C_SCL
    call I2C_DELAY
    bcf I2C_SCL
    call I2C_DELAY
    return
;---------------------
I2C_DEV_ACK:

    bsf I2C_SDAo        ;I2C device acknowledge:
    call I2C_DELAY      ;SDA set 1 by master
    bsf I2C_SCL     ;SDA = 1 if device doesn't ACK
    call I2C_DELAY      ;SDA = 0 if devcie acknowledges

    bsf STATUS,Z        ;STATUS,Z=1 if device acknowledges
    btfsc I2C_SDAi      ;STATUS,Z=0 if device doesn't ACK
    bcf STATUS,Z
    bcf I2C_SCL
    call I2C_DELAY
    return

;---------------------
I2C_SEND:

    movwf I2C_TEMP      ;I2C send 8 bits of working register (W)
                ;to the bus.
    bsf I2C_SDAo
    btfss I2C_TEMP,7
    bcf I2C_SDAo
    call I2C_DELAY
    bsf I2C_SCL
    call I2C_DELAY
    bcf I2C_SCL
    call I2C_DELAY

    bsf I2C_SDAo
    btfss I2C_TEMP,6
    bcf I2C_SDAo
    call I2C_DELAY
    bsf I2C_SCL
    call I2C_DELAY
    bcf I2C_SCL
    call I2C_DELAY

    bsf I2C_SDAo
    btfss I2C_TEMP,5
    bcf I2C_SDAo
    call I2C_DELAY
    bsf I2C_SCL
    call I2C_DELAY
    bcf I2C_SCL
    call I2C_DELAY

    bsf I2C_SDAo
    btfss I2C_TEMP,4
    bcf I2C_SDAo
    call I2C_DELAY
    bsf I2C_SCL
    call I2C_DELAY
    bcf I2C_SCL
    call I2C_DELAY

    bsf I2C_SDAo
    btfss I2C_TEMP,3
    bcf I2C_SDAo
    call I2C_DELAY
    bsf I2C_SCL
    call I2C_DELAY
    bcf I2C_SCL
    call I2C_DELAY

    bsf I2C_SDAo
    btfss I2C_TEMP,2
    bcf I2C_SDAo
    call I2C_DELAY
    bsf I2C_SCL
    call I2C_DELAY
    bcf I2C_SCL
    call I2C_DELAY

    bsf I2C_SDAo
    btfss I2C_TEMP,1
    bcf I2C_SDAo
    call I2C_DELAY
    bsf I2C_SCL
    call I2C_DELAY
    bcf I2C_SCL
    call I2C_DELAY

    bsf I2C_SDAo
    btfss I2C_TEMP,0
    bcf I2C_SDAo
    call I2C_DELAY
    bsf I2C_SCL
    call I2C_DELAY
    bcf I2C_SCL
    call I2C_DELAY

    return

;---------------------
I2C_RECEIVE:

    clrf I2C_TEMP       ;I2C receives 8 bits from the bus
    bsf I2C_SDAo        ;and store them to working register (W)
    call I2C_DELAY

    bsf I2C_SCL
    call I2C_DELAY
    btfsc I2C_SDAi
    bsf I2C_TEMP,7
    bcf I2C_SCL
    call I2C_DELAY

    bsf I2C_SCL
    call I2C_DELAY
    btfsc I2C_SDAi
    bsf I2C_TEMP,6
    bcf I2C_SCL
    call I2C_DELAY

    bsf I2C_SCL
    call I2C_DELAY
    btfsc I2C_SDAi
    bsf I2C_TEMP,5
    bcf I2C_SCL
    call I2C_DELAY

    bsf I2C_SCL
    call I2C_DELAY
    btfsc I2C_SDAi
    bsf I2C_TEMP,4
    bcf I2C_SCL
    call I2C_DELAY

    bsf I2C_SCL
    call I2C_DELAY
    btfsc I2C_SDAi
    bsf I2C_TEMP,3
    bcf I2C_SCL
    call I2C_DELAY

    bsf I2C_SCL
    call I2C_DELAY
    btfsc I2C_SDAi
    bsf I2C_TEMP,2
    bcf I2C_SCL
    call I2C_DELAY

    bsf I2C_SCL
    call I2C_DELAY
    btfsc I2C_SDAi
    bsf I2C_TEMP,1
    bcf I2C_SCL
    call I2C_DELAY

    bsf I2C_SCL
    call I2C_DELAY
    btfsc I2C_SDAi
    bsf I2C_TEMP,0
    bcf I2C_SCL
    call I2C_DELAY
    
    movf I2C_TEMP,0
    return