;*************************************************************************************************
;*************************************************************************************************
;*************************************************************************************************
;;;
;;; (C) 2007,2008 by Fred Bayer <fred@bayerf.de>
;;;
;;; Fred's 4 letter word generator (F4LW)
;;;
;;; Version 1.5 (English/German combined)
;;;
;;; A linguistic sculpture using 4 ancient Burroughs 7971 Nixie tubes
;;; to display random 4 letter words and several other things.
;;;
;;; 29.05.07: Fix delay after editing message
;;; 09.06.07: Daylight saving time, lower/title case for all texts
;;; 10.06.07: Moon phase and sunrise/set times
;;; 17.06.07: Fix displaying **** too early
;;; 17.06.07: Update moon phase twice a day
;;; 06.07.07: Some minor bug fixes
;;; 09.07.07: Added clock synchronization 
;;; 16.07.07: Added small/large clock cycling modes 
;;; 17.07.07: Added 'pendulum' seconds cursor 
;;; 20.07.07: Optionally omit sun table when setting clock
;;; 07.04.08: Bugfix: Init brightness early               
;;; 
;*************************************************************************************************
;;;
;;; IO pin functions:
;;; RA0: tubes data
;;; RA1: tubes strobe cathodes
;;; RA2: tubes strobe anodes
;;; RA3: tubes clock
;;; RA4: ---
;;; RA5: V_pp
;;; RA6: Switch to start upload
;;; RA7: IC SCL
;;; 
;;; RB0: 1 Hz from RTC
;;; RB1: RS232 receive data
;;; RB2: RS232 transmit data
;;; RB3: IC SDAin
;;; RB4: IC SDAout
;;; RB5: IR receive
;;; RB6: ICSP clock
;;; RB7: ICSP data
;;;
;;; Timer0 is used for the IR receiver to measure pulse width
;;; Timer1 is unused
;;; Timer2 is used for multiplexing the display, driven by 1MHz,
;;; scaled by 8*16 = 8kHz resolution, actual update freq. is multiple of this
;;;
;;; 1 Hz signal via RTC, but don't trigger interrupts, just poll INTF
;;; in clock mode
;;;
;;; Battery-backupped RTC is also used to store parameters:
;;; 00-07 : RTC registers
;;; 08    : 0x20 (century)
;;; 09    : DST flag
;;; 0a    : moon-age offset (0-58)
;;; 0b    ; 0x55 (eyecatcher)
;;; 0c    : pstate
;;; 0d    : brightness
;;; 0e    : speed
;;; 0f    ; message number
;;; 10    ; word mode
;;; 11    ; clock mode
;;; 12-39 : 10 messages 4 bytes each
;;;
;*************************************************************************************************
;*************************************************************************************************
;*************************************************************************************************

    list p=16F648a
    #include p16f648a.inc
    __config (_INTRC_OSC_NOCLKOUT & _WDT_OFF & _PWRTE_OFF & _MCLRE_OFF & _CP_OFF & _LVP_OFF)

    ERRORLEVEL -302         ; Ignore register bank warnings

;#define GERMAN (set in Project definition) 0=English, 1=German (different font, weekday names, IR text input)

    cblock 0x20
;----------------------------------------------------------
; Variables in bank 0
;----------------------------------------------------------
num_blocks:2                ; Number of 64 byte blocks to upload
recv_count                  ; counter for bytes to receive on RS232


i2c_adr:2                   ; EEPROM address (big endian)
i2c_byte                    ; Byte sent/received via I2C
i2c_temp                    ; temp for IC routines
eecount                     ; counter for EEPROM buffer
eebufp                      ; pointer into EEPROM buffer
rtc_temp                    ; temporary for RTC access

                            ; Clock variables
                            ; ***************************************
clock_mode                  ; see CM_ flags below, backed up in NV RAM

                            ; 16 bit arithmetic variables
                            ; ***************************************
num:2                       ; 16 bit accumulator
num2:2                      ; 16 bit 2nd operand

                            ; Delay loop counters
                            ; ***************************************
delay1
delay2
delay3

                            ; Random number generator variables
                            ; ***************************************
LSR:6                       ; 47-bit LFSR
RAND                        ; random result
bitcount                    ; also used elsewhere

                            ; Display control variables
                            ; ***************************************
speed                       ; Animation speed, backed up in NV RAM
brightness                  ; Tube brightness, backed up in NV RAM
tube_phase                  ; toggle between duty phases
anodes                      ; anode to be activated; 1 bit shifted right
pattern                     ; pattern to be put into the 4094 shift registers
pcount                      ; counter for pattern bit number
segp                        ; pointer to segment buffer
tube_data:2                 ; 1 bit for each segment on/anode on

buffer:4                    ; ASCII of chars to be displayed
segments:8                  ; Segments (2 bytes per tube) actually displayed
hsegments:8                 ; Hidden segments for animations

                            ; Word generator variables
                            ; ***************************************
word_mode                   ; see WM_* flags below, backed up in NV RAM
num_words:2                 ; Number of words in EEPROM
chain_ptr:2                 ; EEPROM address of word chain
chain_ptr_bak:2             ; backup of chain_ptr
last_chain_ptr:2            ; 2 stage queue to avoid repeating words
last2_chain_ptr:2           ; dito.
word_ptr:2                  ; pointer to word entry
num_trans                   ; number of poss. transitions from current word
diff_char                   ; bit mask indicating which tube has changed
anim_delay                  ; delay in ms for segment animation

                            ; Message variables
                            ; ***************************************
msg_nr                      ; Low 4 bits message nr to display; bit 7 indicates blinking, backed up in NV RAM
edit_tube                   ; index of tube currently being edited

                            ; RC-5 receiver variables
                            ; ***************************************
rc_cmd_rep                  ; received command repeated
rc_cmd                      ; received command not repeated
rc_toggle                   ; bit 0: current toggle bit, bit 1: last toggle bit
rc_bit                      ; edge counter
rc_buf_hi                   ; high byte of received serial IR code
rc_buf_lo                   ; low byte of received serial IR code
rc_state                    ; FSM state: 0: ready, 1: receiving, 3: waiting
    endc

    cblock 0x70 
;----------------------------------------------------------
; RAM shared in all banks
;----------------------------------------------------------
W_temp                      ; Saved W during interrupt
STATUS_temp                 ; Saved STATUS during interrupt
FSR_temp                    ; Saved FSR during interrupt
PCLATH_temp                 ; Saved PCLATH during interrupt

pstate                      ; Program state, backed up in NV RAM, values below
flags                       ; general flags, see FL_* below
wframe                      ; Animation frame counter
anim_pat                    ; Pointer to animation mask pattern (seg_anim)
temp                        ; Local temporary
temp1                       ; Local temporary
temp2                       ; Local temporary
temp3                       ; Local temporary
temp4                       ; Local temporary
temp5                       ; Local temporary
curr_date                   ; Current day of month in binary
curr_month                  ; Current month in binary
    endc

    cblock 0xa0             
;----------------------------------------------------------
; Variables in bank 1
;----------------------------------------------------------
eebuf:.64                   ; 64 byte buffer for uploading EEPROM from RS232
                            ; also used for segment animation!
seg_nr:.14                  ; Array of segment numbers for random animation,
                            ; always a permutation of [0..13]
    endc

;----------------------------------------------------------
; Aliases for shared variables
;----------------------------------------------------------
seg_anim            equ eebuf
curr_year           equ temp5
moon_age            equ temp4

;----------------------------------------------------------
; Bank 2 and 3: unused
;----------------------------------------------------------

;*************************************************************************************************
;*************************************************************************************************
;*
;* IO pins
;*
;*************************************************************************************************
;*************************************************************************************************
#define TUBE_DATA       PORTA,0
#define TUBE_CATHODES   PORTA,1
#define TUBE_ANODES     PORTA,2
#define TUBE_CLOCK      PORTA,3
#define I2C_SDAo        PORTB,4
#define I2C_SDAi        PORTB,3
#define I2C_SCL         PORTA,7
#define RC5_INPUT       PORTB,5
#define BUTTON          PORTA,6

;*************************************************************************************************
;*
;* 24C512 Serial EEPROM via I2C bus 
;*
;*************************************************************************************************
I2C_TEMP            equ i2c_temp
EEPRead             equ 0xa1            ; IC device address to read from EEPROM
EEPWrite            equ 0xa0            ; IC device address to write to EEPROM
RISE_TABLE          equ 0xfa00          ; Start address of sunrise table in EEPROM

;*************************************************************************************************
;*
;* DS1307 real-time clock and NV RAM register numbers 
;*
;*************************************************************************************************
RTCRead             equ 0xd1            ; IC device address to read from RTC
RTCWrite            equ 0xd0            ; IC device address to write to RTC

RTC_SEC             equ 0x00            ; 0 to 59
RTC_MIN             equ 0x01            ; 0 to 59
RTC_HOUR            equ 0x02            ; 0 to 23, always 24h mode
RTC_DAY             equ 0x03            ; of week, 1 (Monday) to 7
RTC_DATE            equ 0x04            ; 1 to 31
RTC_MONTH           equ 0x05            ; 1 to 12
RTC_YEAR            equ 0x06            ; two digits, century in first RAM cell
RTC_CONTROL         equ 0x07            ; RTC control register
RTC_CENTURY         equ 0x08            ; in fact a NV RAM location, always set
                                        ; to 0x20 for easy 4 digit year display
RTC_DST             equ 0x09            ; 0: standard time zone, 1: DST
RTC_MOON            equ 0x0a            ; Moon age offset
RTC_EYE             equ 0x0b            ; Eyecatcher to indicate valid NV RAM
RTC_PSTATE          equ 0x0c            ; copy of pstate
RTC_BRIGHT          equ 0x0d            ; copy of brightness
RTC_SPEED           equ 0x0e            ; copy of speed
RTC_MSG_NR          equ 0x0f            ; copy of msg_nr
RTC_WORD_MODE       equ 0x10            ; copy of word_mode
RTC_CLOCK_MODE      equ 0x11            ; copy of clock_mode
RTC_MESSAGE         equ 0x12            ; Start of message memory

RTC_BLOCKSIZE       equ 0x0b            ; 8 registers + 3 byte NVRAM
RTC_MSGSIZE         equ 4*.10           ; 10 messages 4 bytes each

;*************************************************************************************************
;*
;* Machine states (pstate), keep values in sync with jump table in main loop
;*
;*************************************************************************************************
STATE_TESTING       equ .0              ; To test experimental code
STATE_HELLO         equ .1
STATE_HELLO2        equ .2 
STATE_OFF           equ .3 
STATE_WORDS         equ .4 
STATE_WORD_ANIM     equ .5 
STATE_DOWNLOAD_LEN  equ .6
STATE_DOWNLOAD_ROM  equ .7              ; must be followed by saving state
STATE_SAVE_ROM      equ .8
STATE_CLOCK         equ .9
STATE_DOWNLOAD_TIME equ .10             ; must be followed by saving state
STATE_SAVE_TIME     equ .11
STATE_TERMINAL      equ .12
STATE_MESSAGE       equ .13
STATE_DOWNLOAD_MSG  equ .14             ; must be followed by saving state
STATE_SAVE_MSG      equ .15
STATE_MESSAGE_EDIT  equ .16
STATE_TEST_RC       equ .17
STATE_SHOW_BRIGHT   equ .18
STATE_SHOW_SPEED    equ .19
STATE_SHOW_TAG      equ .20
STATE_CLOCK_SYNC    equ .21

;*************************************************************************************************
;*
;* General flags                                                                 
;*
;*************************************************************************************************
FL_RTC_OK           equ 0               ; RTC has been set
FL_BLINK_ON         equ 1               ; Message blinking: Now display on
FL_ROM_OK           equ 2               ; EEPROM seems valid
FL_HI_BYTE          equ 3               ; Aux. flag for bit masks
FL_DST              equ 4               ; daylight saving time active?

;*************************************************************************************************
;*
;* word_mode flags                                                                 
;*
;*************************************************************************************************
WM_DICT             equ 0               ; Words randomly from dict
WM_CHAIN            equ 1               ; Word chains
WM_70IES            equ 2               ; 70ies emulation
WM_FIRST_LOWER      equ 3               ; First letter to lower case
WM_REST_LOWER       equ 4               ; Other letters to lower case
WM_ROTATE           equ 5               ; Animate rotating segments
WM_SCROLL           equ 6               ; Animate scrolling segments
WM_STICK            equ 7               ; Animate random stick mode

;*************************************************************************************************
;*
;* clock_mode flags (lower nibble contains part to be displayed)
;*
;*************************************************************************************************
CM_SMALL_CYCLE      equ 5               ; 4 common clock displays when set, all 8 when clear 
CM_CYCLE            equ 6               ; Clocks modes are cycling 
CM_SEC_CURSOR       equ 7               ; Display seconds cursor

;*************************************************************************************************
;*
;* Other symbolic constants                                                                 
;*
;*************************************************************************************************
STX                 equ 2               ; ASCII code STX
ETX                 equ 3               ; ASCII code ETX
ACK                 equ 6               ; ASCII code ACK

INC_BRIGHT          equ 1
MAX_BRIGHT          equ .6
INC_SPEED           equ 1
MAX_SPEED           equ .50
ADJ_FRAMES          equ .7              ; Number of frames the adjusted value
                                        ; (speed/bright) is displayed
TAG_FRAMES          equ .15             ; Number of frames ROM tag is displayed

CASE_BIT            equ 5               ; Where lower/upper case chars differ
ALPHA_BIT           equ 6               ; When set, this is (most probably) a alphabetic char
BLINK_MSG           equ 7               ; bit set in msg_nr when blinking
EYE_CATCHER         equ 0x55            ; to indicate valid EEPROM and NV RAM

    #include rc5codes.inc

;*************************************************************************************************
;*************************************************************************************************
;*
;* Macros
;*
;*************************************************************************************************
;*************************************************************************************************

;----------------------------------------------------------
; Select upper 2k segment as goto/call target           
;----------------------------------------------------------
segupper macro
    bsf     PCLATH, 3
    endm

;----------------------------------------------------------
; Select lower 2k segment as goto/call target           
;----------------------------------------------------------
seglower macro
    bcf     PCLATH, 3
    endm

;----------------------------------------------------------
; Test IR button and jump to <Next>, if not pressed
;----------------------------------------------------------
testbut macro Button, Next
    movlw   Button
    subwf   rc_cmd, W
    bnz     Next
    endm

;----------------------------------------------------------
; Test repeating IR button and jump to <Next>, if not pressed
; Otherwise clear button
;----------------------------------------------------------
testrbt macro Button, Next
    movlw   Button                      
    subwf   rc_cmd_rep, W
    bnz     Next
    movlw   0xff
    movwf   rc_cmd_rep
    endm

;----------------------------------------------------------
; Copy 1 byte variable
;----------------------------------------------------------
copy macro From, To
    movfw   From
    movwf   To
    endm

;----------------------------------------------------------
; Copy 2 byte variable
;----------------------------------------------------------
copy2 macro From, To
    movfw   From
    movwf   To
    movfw   From+1
    movwf   To+1
    endm

;----------------------------------------------------------
; Copy 1 byte literal
;----------------------------------------------------------
copyl macro Literal, To
    movlw   Literal
    movwf   To
    endm

;----------------------------------------------------------
; To be used in tube edit mode:
; Test if current char is <Old> and change it to <New>
;----------------------------------------------------------
change macro Old, New                  
    local   _next
    movfw   INDF
    xorlw   Old
    bnz     _next
    movlw   New
    movwf   INDF
    goto    EditDone
_next
    endm

;----------------------------------------------------------
; To be used when filling the display buffer:
; Save char in W to buffer at <Pos> and set corresponding
; changed bit in char_diff, if the saved char differs from
; previous buffer content
;----------------------------------------------------------
savec macro Pos
    movwf   temp2
    xorwf   buffer+Pos, W
    andlw   ~(1<<CASE_BIT)              ; ignore case difference
    skpz
    bsf     diff_char, Pos
    movfw   temp2
    movwf   buffer+Pos
    endm

;----------------------------------------------------------
; To be used for char animation:
; When the char at <CharNr> differs from the previous one,
; apply masking with <temp> and <temp2> bit patterns for
; the segment <SegNr>
;----------------------------------------------------------
smask macro SegNr, CharNr
    movfw   hsegments+SegNr
    btfsc   diff_char, CharNr
    andwf   temp, W
    btfsc   diff_char, CharNr
    iorwf   temp1, W
    movwf   segments+SegNr
    endm

;----------------------------------------------------------
;* Shift 8 bits from Var into shift registers (4094)
;----------------------------------------------------------
outPat macro Var
    local   _ploop
    copyl   0x08, pcount
_ploop
    rrf     Var, F
    skpc
    bcf     TUBE_DATA                   ; 0 on tube data output pin
    skpnc
    bsf     TUBE_DATA                   ; 1 on tube data output pin
    bsf     TUBE_CLOCK                  ; pulse clock
    bcf     TUBE_CLOCK                  ; 1s is enough!
    decfsz  pcount, F
    goto    _ploop
    endm

;----------------------------------------------------------
; Draw a moon phase C1, C2, if moon age is less than <Age>
;----------------------------------------------------------
drawMoon macro Age, C1, C2
    local   _next
    movlw   Age
    subwf   moon_age, W
    bc      _next
    copyl   C1, buffer+2   
    copyl   C2, buffer+3
    goto    _moonDone
_next
    endm

;*************************************************************************************************
;*************************************************************************************************
;*************************************************************************************************
;*
;* CODE STARTS HERE --- CODE STARTS HERE --- CODE STARTS HERE --- CODE STARTS HERE --- CODE STARTS
;*
;*************************************************************************************************
;*************************************************************************************************
;*************************************************************************************************
;*
;* LOWER 2k PAGE --- LOWER 2k PAGE --- LOWER 2k PAGE --- LOWER 2k PAGE --- LOWER 2k PAGE --- LOWER
;*
;*************************************************************************************************
;*************************************************************************************************
;*************************************************************************************************

    org     0                           ; Reset vector
    copyl   0x07, CMCON                 ; turn comparators off
    clrwdt                              ; Clear the Watchdog timer
    goto    Setup

;*************************************************************************************************
;*************************************************************************************************
;*
;* Interrupt handler (continued in upper page)
;*
;*************************************************************************************************
;*************************************************************************************************
    org     4                           ; Interrupt vector
    movwf   W_temp                      ; Standard prolog: 
    swapf   STATUS, W                   ; save W, STATUS and PCLATH
    movwf   STATUS_temp
    bcf     STATUS, RP0
    copy    PCLATH, PCLATH_temp
    segupper                            ; Continue in upper 2k block
    goto    IRHandler           

;*************************************************************************************************
;*************************************************************************************************
;*************************************************************************************************
;*
;* Hardware setup and subsystem initialization
;*
;*************************************************************************************************
;*************************************************************************************************
;*************************************************************************************************
Setup
    clrf    word_mode                   ; to not convert case on NV RAM init
    copyl   b'10000000', anodes         ; Init current anode for interrupt
    copyl   segments, segp              ; Init pointer to segments
    copyl   MAX_BRIGHT, brightness      ; Must be defined before enabling interrupts
    copyl   b'11101000', INTCON         ; Global, peripheral, RB pin change, and Timer0 interrupts
    bsf     STATUS, RP0                 ; Bank 1
    copyl   b'01110000', TRISA          ; RA7, RA0-RA3 output
    copyl   b'11101111', TRISB          ; RB1 and RB2 must be both 1 to enable
                                        ; serial port, RB4 output for 
                                        ; I2C emulation
    copyl   b'00000010', OPTION_REG     ; Timer 0 prescale 1:8, internal,
                                        ; WPU enable, INT on falling edge
    copyl   b'00100010', PIE1           ; Enable serial port receive and
                                        ; Timer2 interrupts
    bcf     STATUS, RP0                 ; Bank 0
    copyl   b'00111111', T2CON          ; Prescale 1:16, Postscale 1:8

    copyl   0xff, rc_cmd                ; -1: No command received
    movwf   rc_cmd_rep                  ; -1: No command received
    clrf    rc_state                    ; waiting for IR edge
    clrf    flags

    call    ClearSegments               ; Blank display 
    call    SerInit
    call    RTCInit
    call    LoadNVParms
    call    RandomInit
    call    WordInit
    copyl   STATE_HELLO, pstate         ; Display welcome message
    
;*************************************************************************************************
;*************************************************************************************************
;*************************************************************************************************
;*
;* Main Loop ---- Main Loop ---- Main Loop ---- Main Loop ---- Main Loop ---- Main Loop ---- Main
;*
;*************************************************************************************************
;*************************************************************************************************
;*************************************************************************************************
MainLoop
    copyl   HIGH MainLoop, PCLATH
    movfw   pstate
    addwf   PCL, F
    goto    StateTesting
    goto    StateHello
    goto    StateHello2
    goto    StateOff
    goto    StateWords
    goto    StateWordAnim
    goto    StateDownloadLen
    goto    StateDownloadRom
    goto    StateSaveRom
    goto    StateClock
    goto    StateDownloadTime
    goto    StateSaveTime
    goto    StateTerminal
    goto    StateMessage
    goto    StateDownloadMsg
    goto    StateSaveMsg
    goto    StateMessageEdit
    goto    StateTestRc
    goto    StateShowBright
    goto    StateShowSpeed
    goto    StateShowTag
    goto    StateClockSync

;----------------------------------------------------------
;* Insert your testing code here
;----------------------------------------------------------
StateTesting
    movlw   .1
    call    delayTenth
    goto    MainLoop

;----------------------------------------------------------
;* All tubes off                          
;----------------------------------------------------------
StateOff
    call    ClearSegments
    movlw   .1
    call    delayTenth
    goto    CheckIR

;----------------------------------------------------------
;* Display hello message
;----------------------------------------------------------
StateHello
    movlw   .17                         ; "F4LW"
    call    DisplayWord
    movlw   .15                         ; 1.5s delay
    call    delayTenth
    copyl   STATE_HELLO2, pstate        ; next, display version
    goto    MainLoop

;----------------------------------------------------------
;* Display version string                            
;----------------------------------------------------------
StateHello2
    movlw   .15                         ; Version string in internal EEPROM
    call    DisplayWord
    movlw   .15                         ; 1.5s delay
    call    delayTenth
    btfss   BUTTON                      ; Test download button
    goto    _startRc

    movlw   RTC_PSTATE                  ; Restore pstate from NV RAM
    call    ReadRtc
    movwf   pstate
    goto    NewState                    ; and continue there

_startRc
    copyl   STATE_TEST_RC, pstate       ; but don't store in NV RAM
    goto    MainLoop

;----------------------------------------------------------
;* Display random 4 letter words                           
;----------------------------------------------------------
StateWords 
    btfss   BUTTON                      ; Test download button
    goto    StartDownload

    decfsz  wframe, F
    goto    _keepWord

    call    InitWFrame
    btfsc   word_mode, WM_DICT          ; Select from entire dictionary
    call    SelectWord
    btfsc   word_mode, WM_CHAIN         ; Select word differing by one character
    call    SelectChainedWord
    btfsc   word_mode, WM_70IES         ; Emulate original 70ies 4-letter-word
    call    Make70iesWord

    call    HandleCase

    ;-----------------------------------
    ; Handle animation initialization
    ;-----------------------------------
    btfsc   word_mode, WM_ROTATE
    goto    InitRotate
    btfsc   word_mode, WM_SCROLL
    goto    InitScroll
    btfsc   word_mode, WM_STICK
    goto    InitStick
    copyl   .100, anim_delay
    call    ConvertSegments

_keepWord                               ; nothing new to display
    movlw   .100
    call    delayMs
    
    ;-----------------------------------
    ; Check remote control for word mode
    ; options (keys 0-9)
    ;-----------------------------------
CheckWordOptions
    testbut BTN_DIGIT_1, _c41           ; Random from dictionay
    bcf     word_mode, WM_70IES
    bcf     word_mode, WM_CHAIN
    bsf     word_mode, WM_DICT
    goto    _newWordMode
_c41
    testbut BTN_DIGIT_2, _c42           ; Word chain
    bcf     word_mode, WM_70IES
    bsf     word_mode, WM_CHAIN
    bcf     word_mode, WM_DICT
    goto    _newWordMode
_c42
    testbut BTN_DIGIT_3, _c43           ; 70ies emulation
    bsf     word_mode, WM_70IES
    bcf     word_mode, WM_CHAIN
    bcf     word_mode, WM_DICT
    clrf    chain_ptr                   ; Clear word chain: no previous words
    clrf    chain_ptr+1                 ; to avoid
    goto    _newWordMode
_c43
    testbut BTN_DIGIT_4, _c44           ; All upper case
    bcf     word_mode, WM_FIRST_LOWER
    bcf     word_mode, WM_REST_LOWER
    goto    _newWordMode
_c44
    testbut BTN_DIGIT_5, _c45           ; Title case
    bcf     word_mode, WM_FIRST_LOWER
    bsf     word_mode, WM_REST_LOWER
    goto    _newWordMode
_c45
    testbut BTN_DIGIT_6, _c46           ; All lower case
    bsf     word_mode, WM_FIRST_LOWER
    bsf     word_mode, WM_REST_LOWER
    goto    _newWordMode
_c46
    testbut BTN_DIGIT_7, _c47           ; Scroll down
    bcf     word_mode, WM_ROTATE
    bsf     word_mode, WM_SCROLL
    bcf     word_mode, WM_STICK
    goto    _newWordMode
_c47
    testbut BTN_DIGIT_8, _c48           ; Rotating segments
    bsf     word_mode, WM_ROTATE
    bcf     word_mode, WM_SCROLL
    bcf     word_mode, WM_STICK
    goto    _newWordMode
_c48
    testbut BTN_DIGIT_9, _c49           ; Random segments
    bcf     word_mode, WM_ROTATE
    bcf     word_mode, WM_SCROLL
    bsf     word_mode, WM_STICK
    goto    _newWordMode
_c49
    testbut BTN_DIGIT_0, _c50           ; No animation
    bcf     word_mode, WM_ROTATE
    bcf     word_mode, WM_SCROLL
    bcf     word_mode, WM_STICK
    goto    _newWordMode
_c50
    testbut BTN_MENU, CheckIR           ; Display ROM tag
    copyl   STATE_SHOW_TAG, pstate
    goto    IRDone

_newWordMode
    copyl   RTC_WORD_MODE, FSR          ; Also save to NV RAM
    movfw   word_mode
    call    WriteRtc
    copyl   1, wframe                   ; Immediately terminate waiting
    goto    IRDone

StartDownload
    ;-----------------------------------
    ; Start downloading EEPROM, first
    ; 2 bytes number of blocks:
    ;-----------------------------------
    copyl   .2, recv_count
    copyl   STATE_DOWNLOAD_LEN, pstate
    goto    MainLoop

;----------------------------------------------------------
;* Init table for random stick animation: shuffle segment
;* numbers and build bit masks               
;----------------------------------------------------------
InitStick
    ;-----------------------------------
    ; Use Knuth shuffle for random
    ; permutation of seg_nr array
    ;-----------------------------------
    copyl   .14, num                    ; can't use bitcount here (RandomRange!)
    copyl   seg_nr, temp2               ; temp2 is ptr to top item
_swapSeg
    movfw   num
    call    RandomRange                 ; uses temp and temp1
    addwf   temp2, W
    movwf   temp3                       ; temp3 is ptr to random item after top
    movwf   FSR                         ; temp and temp1 usable again
    copy    INDF, temp                  ; swap contents of temp2 and temp3
    copy    temp2, FSR                  ; via temp and temp1
    copy    INDF, temp1                 
    copy    temp, INDF
    copy    temp3, FSR
    copy    temp1, INDF
    incf    temp2, F                    ; advance top
    decfsz  num, F                      ; gone thru entire stack?
    goto    _swapSeg

    ;-----------------------------------
    ; Accumulate bit masks in num,
    ; store them to num2
    ;-----------------------------------
    clrf    num
    clrf    num+1
    copyl   .14, bitcount               ; Number of animation frames
    movwf   wframe
    copyl   seg_anim, temp3             ; Ptr to mask buffer
    copyl   seg_nr, temp4               ; Ptr to current segment nr
_nextSegment    
    bcf     flags, FL_HI_BYTE
    copy    temp4, FSR
    btfsc   INDF, 3                     ; Test, if hi or lo byte
    bsf     flags, FL_HI_BYTE
    copyl   HIGH BitShift, PCLATH       ; BitShift uses jump table
    movfw   INDF
    andlw   0x07
    call    BitShift                    ; Make mask from bit nr.
    btfsc   flags, FL_HI_BYTE
    goto    _highByte
    clrf    num2+1
    movwf   num2                        ; Flash segment
    iorwf   num, F                      ; Accumulate mask to hi byte
    goto    _storeMask
_highByte
    clrf    num2
    movwf   num2+1                      ; Flash segment
    iorwf   num+1, F                    ; Accumulate mask to lo byte
_storeMask
    copy    temp3, FSR                  
    copy    num, INDF                   ; Store hi mask to buffer
    incf    FSR, F
    copy    num+1, INDF                 ; Store lo mask to buffer
    incf    FSR, F
    copy    num2, INDF                  ; Store hi flashing segment
    incf    FSR, F
    copy    num2+1, INDF                ; Store lo flashing segment
    incf    FSR, F
    copy    FSR, temp3                  ; Points to next mask block
    incf    temp4, F                    ; Next segment nr
    decfsz  bitcount, F
    goto    _nextSegment                    
    copyl   .60, anim_delay             ; 60ms for each phase
    goto    FinishAnimInit

;----------------------------------------------------------
;* Init table for rotating segment animation: just copy
;* from EEPROM table                          
;----------------------------------------------------------
InitRotate
    copyl   LOW rotatePat, FSR
    movlw   scrollPat-rotatePat
    call    CopyAnim
    copyl   (scrollPat-rotatePat)/4, wframe ; Number of animation frames
    copyl   .80, anim_delay             ; 80ms for each phase
    goto    FinishAnimInit

;----------------------------------------------------------
;* Init table for scrolling animation: just copy from
;* EEPROM table                           
;----------------------------------------------------------
InitScroll
    copyl   LOW scrollPat, FSR
    movlw   lastPat-scrollPat
    call    CopyAnim
    copyl   (lastPat-scrollPat)/4, wframe   ; Number of animation frames
    copyl   .120, anim_delay            ; 120ms for each phase

FinishAnimInit
    ;-----------------------------------
    ; Common continuation 
    ;-----------------------------------
    call    ConvertHiddenSegments
    copyl   seg_anim, anim_pat
    copyl   STATE_WORD_ANIM, pstate
    goto    MainLoop

;----------------------------------------------------------
;* Display word animation   
;----------------------------------------------------------
StateWordAnim
    ;-----------------------------------
    ; Animate LO segments if changed
    ;-----------------------------------
    copy    anim_pat, FSR               ; offset 0
    copy    INDF, temp                  ; LO of and mask
    incf    anim_pat, F
    incf    anim_pat, F
    copy    anim_pat, FSR               ; offset 2
    copy    INDF, temp1                 ; LO of or mask

    smask   1, 0                         
    smask   3, 1
    smask   5, 2
    smask   7, 3

    ;-----------------------------------
    ; Animate HI segments if changed
    ;-----------------------------------
    decf    anim_pat, F
    copy    anim_pat, FSR               ; offset 1
    copy    INDF, temp                  ; HI of and mask
    incf    anim_pat, F
    incf    anim_pat, F
    copy    anim_pat, FSR               ; offset 3
    copy    INDF, temp1                 ; HI of or mask

    smask   0, 0
    smask   2, 1
    smask   4, 2
    smask   6, 3

    ;-----------------------------------
    ; next animation step
    ;-----------------------------------
    incf    anim_pat, F                 ; offset 4 => next animation masks
    decfsz  wframe, F
    goto    _animDelay

    call    ConvertSegments
    call    InitWFrame                  ; Animation complete
    copyl   STATE_WORDS, pstate
    goto    MainLoop                 

_animDelay
    movfw   anim_delay
    call    delayMs
    goto    CheckWordOptions            ; now check for RC commands

;----------------------------------------------------------
;* Display number of remaining blocks to be downloaded   
;----------------------------------------------------------
StateDownloadRom
    copy2   num_blocks, num
    segupper    
    call    ConvertDecimal
    seglower    
    call    ConvertSegments
    goto    MainLoop

;----------------------------------------------------------
;* Display prompt to start EEPROM download   
;----------------------------------------------------------
StateDownloadLen
    movlw   8                           ; ">UW<"    
    call    DisplayWord
    goto    MainLoop

;----------------------------------------------------------
;* Save data block from eebuf to EEPROM                    
;----------------------------------------------------------
StateSaveRom
    call    WriteEEBuf
    ;-----------------------------------
    ; advance write address
    ;-----------------------------------
    movlw   .64
    addwf   i2c_adr+1, F
    skpnc
    incf    i2c_adr, F

    movlw   ACK
    call    SerSend                     ; notify PC to send next block

    ;-----------------------------------
    ; check for more blocks to receive
    ;-----------------------------------
    movlw   1                           ; 16 bit decrement and zero test :-(
    subwf   num_blocks+1, F
    skpc
    decf    num_blocks, F
    movf    num_blocks, F
    bnz     _moreBlocks
    movf    num_blocks+1, F
    bnz     _moreBlocks

    ;-----------------------------------
    ; all blocks received
    ;-----------------------------------
AllBlocks
    movlw   .10                         ; display " OK "
    call    DisplayWord
    movlw   RTC_PSTATE                  ; Restore pstate from NV RAM
    call    ReadRtc
    movwf   pstate
    movlw   .20                         ; for 2s
    call    delayTenth
    call    WordInit
    goto    MainLoop

_moreBlocks
    copyl   eebuf, eebufp
    copyl   .64, recv_count
    copyl   STATE_DOWNLOAD_ROM, pstate
    goto    MainLoop

;----------------------------------------------------------
;* Display current time                                    
;----------------------------------------------------------
StateClock
    btfss   BUTTON                      ; Test download button
    goto    _startSetClock

    btfsc   flags, FL_RTC_OK
    goto    _checkCycle
    movlw   .16                         ; "RTC?"
    call    DisplayWord
    movlw   .1
    call    delayTenth
    goto    CheckIR

_checkCycle
    btfss   clock_mode, CM_CYCLE        ; Check for cycling display
    goto    _check1Hz 
    movlw   .1
    call    delayTenth
    decfsz  wframe, F                   ; Next display?
    goto    _check1Hz 
    incf    clock_mode, F
    bcf     clock_mode, 3               ; to keep clock_mode < 8
    btfsc   clock_mode, CM_SMALL_CYCLE  ; Small cycle?
    bcf     clock_mode, 2               ; keep clock_mode < 4
    call    InitWFrame
    goto    _needUpdate
_check1Hz
    btfss   INTCON, INTF                ; 1 Hz interrupt from RTC?
    goto    _noUpdate
_needUpdate
    call    UpdateClock
    bcf     INTCON, INTF                ; re-enable  interrupts

_noUpdate
    ;-----------------------------------
    ; Now check for RC commands
    ;-----------------------------------
    testbut BTN_MENU, _b41              ; Seconds cursor
    movfw   clock_mode                  ; Toggle seconds cursor bit
    xorlw   1<<CM_SEC_CURSOR
    goto    _newClockMode
_b41
    testbut BTN_DIGIT_9, _b42           ; Large cycle display
    call    InitWFrame                  ; Reset animation frame counter
    movlw   1<<CM_CYCLE
    goto    _maskClockMode
_b42
    testbut BTN_DIGIT_0, _b43           ; Small cycle display
    call    InitWFrame                  ; Reset animation frame counter
    movlw   1<<CM_CYCLE | 1<<CM_SMALL_CYCLE
    goto    _maskClockMode
_b43
    testbut BTN_ARITY, _b44             ; Toggle DST
    movlw   RTC_HOUR                    ; Read hour from RTC 
    call    ReadRtc
    call    BCD2Bin
    btfss   flags, FL_DST
    goto    DstOn

    ;-----------------------------------
    ; Switch DST off
    ;-----------------------------------
    addlw   -1                          ; 1 hour back
    bnc     WrapAround                  ; would wrap before midnight?
    call    Bin2BCD
    movwf   temp
    copyl   RTC_HOUR, FSR
    movfw   temp
    call    WriteRtc                    ; Update hour in RTC
    copyl   RTC_DST, FSR
    clrw
    call    WriteRtc                    ; Clear DST flag in RTC
    bcf     flags, FL_DST
    goto    _refreshDisplay

    ;-----------------------------------
    ; Switch DST on
    ;-----------------------------------
DstOn
    addlw   1                           ; advance 1 hour
    movwf   temp
    sublw   .23
    bnc     WrapAround                  ; would wrap after midnight?
    movfw   temp
    call    Bin2BCD
    movwf   temp
    copyl   RTC_HOUR, FSR
    movfw   temp                        ; Update hour in RTC
    call    WriteRtc
    copyl   RTC_DST, FSR
    movlw   1
    call    WriteRtc                    ; Set DST flag in RTC 
    bsf     flags, FL_DST
    goto    _refreshDisplay

WrapAround
    ;-----------------------------------
    ; Wrap-around to other day, forbid this
    ;-----------------------------------
    movlw   .24                         ; "WRAP"
    call    DisplayWord
    movlw   .5
    call    delayTenth
    goto    IRDone

_b44
    movfw   rc_cmd                      ; Keys 0-8
    sublw   .8
    bnc     CheckIR                     ; too big?
    movf    rc_cmd, F                   ; Key 0?
    bz     _b42
_b45
    decf    rc_cmd, W                   ; Store keycode-1 as display mode

_maskClockMode                          ; Set low order bits but keep cursor flag
    movwf   temp
    movfw   clock_mode
    andlw   1<<CM_SEC_CURSOR
    iorwf   temp, W

_newClockMode
    movwf   clock_mode
    copyl   RTC_CLOCK_MODE, FSR         ; Also save to NV RAM
    movfw   clock_mode
    call    WriteRtc
_refreshDisplay
    bsf     INTCON, INTF                ; Simulate 1 Hz interrupt to force redisplay
    goto    IRDone

_startSetClock
    movfw   clock_mode                  ; Test for clock sync vs. download time
    andlw   b'00001111'                 ; Ignore cycle and second cursor flag
    xorlw   .1                          ; Clock mode == mm:ss ?
    bz      _startClockSync  
    copyl   eebuf, eebufp
    copyl   RTC_BLOCKSIZE, recv_count
    copyl   STATE_DOWNLOAD_TIME, pstate
    goto    MainLoop

_startClockSync
    copyl   STATE_CLOCK_SYNC, pstate
    copyl   0xff, rc_cmd
    goto    MainLoop

;----------------------------------------------------------
;* Wait for timedata block on RS232                     
;----------------------------------------------------------
StateDownloadTime
    movlw   .9                          ; ">SC<"
    call    DisplayWord
    goto    MainLoop

;----------------------------------------------------------
;* Copy received time block to RTC and acknowledge
;----------------------------------------------------------
StateSaveTime
    clrf    FSR                         ; Start address 0
    copyl   RTC_BLOCKSIZE, eecount      ; Number of bytes to write
    call    WriteClock
    bcf     flags, FL_DST
    call    RTCInit
    btfss   temp2, 7                    ; Set when need to receive sun table
    goto    _timeDone

    ;-----------------------------------
    ; Now prepare for receiving
    ; sunrise/set table
    ;-----------------------------------
    clrf    num_blocks
    copyl   .23, num_blocks+1           ; 23*64 = 1472 > 1460 = 4*366
    copyl   HIGH RISE_TABLE, i2c_adr    ; Startaddress in EEPROM of table
    clrf    i2c_adr+1                   ; This wastes 64 bytes at the EEPROM top, so what
    copyl   eebuf, eebufp
    copyl   .64, recv_count
    copyl   STATE_DOWNLOAD_ROM, pstate
    movlw   ACK                         ; Acknowledge time block 
    call    SerSend
    goto    MainLoop

_timeDone
    movlw   ACK                         ; Acknowledge time block 
    call    SerSend
    goto    AllBlocks                   ; Finished

;----------------------------------------------------------
;* Manually synchronize time with external clock
;----------------------------------------------------------
StateClockSync
    movlw   RTC_MIN<<4 | RTC_SEC        ; Display mm:ss
    call    DispRegister
    movlw   b'11110000'
    call    DrawCursor                  ; and set all cursor segments, too
    movlw   .1
    call    delayTenth

    ;-----------------------------------
    ; Set to nearest minute on ENTER press
    ;-----------------------------------
    testbut BTN_ENTER, IRDone
    movlw   RTC_SEC
    call    ReadRtc                     ; Test RTC seconds
    sublw   0x30
    bc      _zeroSec                    ; > 30s? => increment minutes

    copyl   RTC_MIN, FSR                ; Read RTC minutes
    call    ReadRtc
    call    BCD2Bin
    addlw   1                           ; Increment minutes
    call    Bin2BCD
    movwf   temp
    sublw   0x59                        ; Check for overflow to next hour 
    bnc     _incrHour
    movfw   temp    
    call    WriteRtc                    

_zeroSec
    copyl   RTC_SEC, FSR                ; Clear RTC seconds
    clrw
    call    WriteRtc
    
    copyl   STATE_CLOCK, pstate
    goto    IRDone

_incrHour
    copyl   RTC_HOUR, FSR               ; Read RTC hours
    call    ReadRtc
    call    BCD2Bin
    addlw   1                           ; Increment hours
    call    Bin2BCD
    movwf   temp
    sublw   0x23                        ; Check for overflow to next day
    bnc     _wrap                       ; not allowed!
    movfw   temp    
    call    WriteRtc                    
    copyl   RTC_MIN, FSR                ; Clear RTC minutes
    clrw
    call    WriteRtc
    goto    _zeroSec 

_wrap
    copyl   STATE_CLOCK, pstate
    goto    WrapAround 

;----------------------------------------------------------
;* Receive characters from PC and display them
;----------------------------------------------------------
StateTerminal
    call    ConvertSegments             ; Buffer has been filled in interrupt,
                                        ; so just display
    ;-----------------------------------
    ; Convert input from RC and
    ; send to PC
    ;-----------------------------------
    movfw   rc_cmd
    sublw   0x60                        ; Colored buttons > 0x60, let system
    bnc     CheckIR                     ; handle them
    segupper
    call    MapIRKey                    ; convert to char to be sent to PC
    seglower
    addlw   0                           ; To check for Z flag
    bz      CheckIR
    call    SerSend                     ; Send translated IR command to PC
    copyl   0xff, rc_cmd_rep
    goto    IRDone

;----------------------------------------------------------
;* Message mode                          
;----------------------------------------------------------
StateMessage
    btfss   BUTTON                      ; Test download button
    goto    StartDownloadMsg

    decfsz  wframe, F
    goto    _keepDisp
    call    InitWFrame
    movfw   flags                       ; Toggle blink-on flag 
    xorlw   1<<FL_BLINK_ON
    movwf   flags

    call    LoadRtcWord                 ; Load message
    btfss   msg_nr, BLINK_MSG
    goto    _convSeg                    ; not blinking? -> display
    btfsc   flags, FL_BLINK_ON
    goto    _convSeg                    ; blinking, but currently on? -> display
    call    ClearSegments
    goto    _keepDisp
_convSeg
    call    ConvertSegments
_keepDisp
    movlw   .1
    call    delayTenth

    ;-----------------------------------
    ; Check RC commands
    ;-----------------------------------
    testbut BTN_ENTER, _m0              ; Start editing current word
    copyl   STATE_MESSAGE_EDIT, pstate
    copyl   0, edit_tube                ; begin with leftmost tube
    goto    IRDone              
_m0
    testbut BTN_MENU, _m1               ; Toggle blinking
    movfw   msg_nr
    xorlw   1<<BLINK_MSG
    goto    _saveNr

_m1
    movfw   rc_cmd                      ; Select a predefined message 0-9
    sublw   .9
    bnc     CheckIR                     ; > 9? 
    movfw   msg_nr
    andlw   1<<BLINK_MSG                ; Keep blinking bit
    iorwf   rc_cmd, W                   ; set message number
_saveNr
    movwf   msg_nr
    copyl   RTC_MSG_NR, FSR             ; Also save to NV RAM
    movfw   msg_nr
    call    WriteRtc
ForceRedisplay
    copyl   1, wframe                   ; Force new word to appear in next frame
    bcf     flags, FL_BLINK_ON          ; set phase off, so it will appear now   
    goto    IRDone

StartDownloadMsg
    copyl   eebuf, eebufp
    copyl   RTC_MSGSIZE, recv_count
    copyl   STATE_DOWNLOAD_MSG, pstate
    goto    MainLoop

;----------------------------------------------------------
;* Wait for message block on RS232                     
;----------------------------------------------------------
StateDownloadMsg
    movlw   .14                         ; ">UM<"
    call    DisplayWord
    goto    MainLoop

;----------------------------------------------------------
;* Copy received message block to NV RAM and acknowledge
;----------------------------------------------------------
StateSaveMsg
    copyl   RTC_MESSAGE, FSR            ; Start address
    copyl   RTC_MSGSIZE, eecount        ; Number of bytes to write
    call    WriteClock

    movlw   ACK
    call    SerSend
    movlw   .10                         ; " OK "
    call    DisplayWord
    copyl   STATE_MESSAGE, pstate
    movlw   .20
    call    delayTenth
    goto    MainLoop

;----------------------------------------------------------
;* Editing the currently displayed message
;----------------------------------------------------------
StateMessageEdit
    call    ConvertSegments
    movlw   segments
    addwf   edit_tube, W
    addwf   edit_tube, W
    movwf   FSR
    bsf     INDF, 1                     ; Set cursor segment
    movlw   .2
    call    delayTenth

    testbut BTN_CURS_LEFT, _e1          ; Edit previous tube
    decf    edit_tube, W
    andlw   0x03
    movwf   edit_tube
    goto    IRDone
_e1
    testbut BTN_CURS_RIGHT, _e2         ; Edit next tube
    incf    edit_tube, W
    andlw   0x03
    movwf   edit_tube
    goto    IRDone
_e2
    testbut BTN_ENTER, _e3              ; Terminate editing and save
    call    SaveRtcWord
    copyl   STATE_MESSAGE, pstate
    goto    ForceRedisplay
_e3
    testbut BTN_BACK, TestRcKeys        ; Abort editing
    copyl   STATE_MESSAGE, pstate
    goto    ForceRedisplay

TestRcKeys
    ;-----------------------------------
    ; Check RC command to edit
    ; current tube
    ;-----------------------------------
    segupper
    call    EditTube
    seglower
    goto    MainLoop

;----------------------------------------------------------
;* Display the received IR code
;----------------------------------------------------------
StateTestRc
    movlw   .20                         ; "RC  "
    call    LoadWord
    movfw   rc_cmd
    segupper
    call    ConvertHex
    seglower
    call    ConvertSegments
    btfsc   rc_toggle, 0                ; Display toggle bit as cursor segment
    bsf     segments+6, 1
    movlw   .1
    call    delayTenth
    goto    MainLoop                    ; We can't break out via IR!

;----------------------------------------------------------
;* Temporarily show brightness
;----------------------------------------------------------
StateShowBright
    movlw   .19                         ; "BR  "
    call    LoadWord
    movfw   brightness
_showTemp1
    call    Bin2BCD
    setc                                ; display in right half
    call    ConvertBCD
    copyl   ADJ_FRAMES, wframe
_showTemp2
    call    ConvertSegments
    movlw   RTC_PSTATE                  ; Restore previous pstate
    call    ReadRtc
    movwf   pstate
    goto    CheckIR

;----------------------------------------------------------
;* Temporarily show speed
;----------------------------------------------------------
StateShowSpeed
    movlw   .18                         ; "SP  "
    call    LoadWord
    movfw   speed
    goto    _showTemp1

;----------------------------------------------------------
;* Temporarily show ROM tag
;----------------------------------------------------------
StateShowTag
    call    FetchROMTag
    copyl   TAG_FRAMES, wframe
    goto    _showTemp2

;*************************************************************************************************
;*
;* Check IR code and determine next major state 
;*
;*************************************************************************************************
CheckIR
    call    Brightness
    testbut BTN_OFF, _b0
    copyl   STATE_OFF, pstate
    goto    NoSaveNvRam                 ; After powering-on, we want something to display, not blank!
_b0
    testbut BTN_GREEN, _b1
    movlw   STATE_CLOCK
    goto    NewState
_b1
    testbut BTN_YELLOW, _b2
    movlw   STATE_TERMINAL
    goto    NewState
_b2 
    testbut BTN_RED, _b3
    movlw   STATE_WORDS
    goto    NewState
_b3
    testbut BTN_BLUE, MainLoop
    movlw   STATE_MESSAGE

NewState
    ;-----------------------------------
    ; Store new pstate to NV RAM, too
    ;-----------------------------------
    movwf   pstate
    copyl   RTC_PSTATE, FSR
    movfw   pstate
    call    WriteRtc

NoSaveNvRam
    ;-----------------------------------
    ; pstate has changed, notify PC to
    ; handle terminal server 
    ;-----------------------------------
    copyl   1, wframe
    movfw   pstate
    xorlw   STATE_TERMINAL
    bz      _startTerminal
    movlw   ETX                         ; Another mode selected, notify PC
    call    SerSend
    goto    IRDone
_startTerminal
    movlw   .21
    call    DisplayWord                 ; "SER?", shifted out when PC responds
    movlw   STX                         ; Notify PC to start Terminal mode
    call    SerSend
IRDone
    copyl   0xff, rc_cmd
    goto    MainLoop

;*************************************************************************************************
;*
;* Update clock display
;*
;*************************************************************************************************
UpdateClock
    ;-----------------------------------
    ; Prefetch date, month, year for 
    ; sun and moon calculations 
    ;-----------------------------------
    movlw   RTC_MONTH
    call    ReadRtc
    call    BCD2Bin    
    movwf   curr_month
    movlw   RTC_DATE
    call    ReadRtc
    call    BCD2Bin    
    movwf   curr_date

    copyl   HIGH UpdateClock, PCLATH    ; to care for jump table
    movfw   clock_mode
    andlw   0x1f                        ; Mask out cycle and seconds cursor bits
    addwf   PCL, F
    goto    _cmHHMM                     ; First 4 modes in small cycle
    goto    _cmMMSS
    goto    _cmDDMM
    goto    _cmWeekDay
    goto    _cmCCYY                     ; Large cycle begins here 
    goto    _cmMoonPhase
    goto    _cmSunRiseSet               ; Rise and set handled by same routine!
    goto    _cmSunRiseSet

_cmHHMM
    movlw   RTC_HOUR<<4 | RTC_MIN
    goto    DispRegister
_cmMMSS
    movlw   RTC_MIN<<4 | RTC_SEC
    goto    DispRegister
_cmDDMM
    movlw   RTC_DATE<<4 | RTC_MONTH
    goto    DispRegister
_cmCCYY
    movlw   RTC_CENTURY<<4 | RTC_YEAR

DispRegister
    ;-----------------------------------
    ; Display 2 RTC registers, whose 
    ; numbers are in upper/lower nibble
    ; of W
    ;-----------------------------------
    movwf   temp2
    swapf   temp2, W                    ; Read RTC register in upper nibble 
    andlw   0x0f
    call    ReadRtc
    clrc
    call    ConvertBCD                  ; and display it on left tubes

    movf    temp2, W                    ; Read RTC register in lower nibble 
    andlw   0x0f
    call    ReadRtc
    setc
    call    ConvertBCD                  ; and display it on right tubes
    call    ConvertSegments

_testCursor
    ;-----------------------------------
    ; have to display seconds cursor?
    ;-----------------------------------
    btfss   clock_mode, CM_SEC_CURSOR
    return
    movlw   RTC_SEC                     ; read lo nibble
    call    ReadRtc
    segupper
    call    MakeSecCursor
    seglower

DrawCursor
    ;-----------------------------------
    ; Light cursor segment, if corr. bit
    ; set in W
    ;-----------------------------------
    movwf   temp
    btfsc   temp, 7
    bsf     segments, 1                 ; add cursor segment
    btfsc   temp, 6
    bsf     segments+2, 1               ; add cursor segment
    btfsc   temp, 5
    bsf     segments+4, 1               ; add cursor segment
    btfsc   temp, 4
    bsf     segments+6, 1               ; add cursor segment
    return

_cmWeekDay
    ;-----------------------------------
    ; Display day of week and DST flag 
    ;-----------------------------------
    movlw   RTC_DAY
    call    ReadRtc
    call    LoadWord    
    movlw   '*'
    btfsc   flags, FL_DST               ; indicate DST
    movwf   buffer+3
_drawIt
    call    ConvertSegments
    goto    _testCursor

_cmMoonPhase
    ;-----------------------------------
    ; Display moon phase graphically 
    ;-----------------------------------
    movlw   .22                         ; "MP  "
    call    LoadWord
    segupper
    call    MoonPhase
    seglower
    goto    _drawIt

_cmSunRiseSet
    ;-----------------------------------
    ; Compute address of sunrise/set table
    ; entry for current date into i2c_adr
    ;-----------------------------------
    clrf    i2c_adr
    clrf    temp2                       ; Correction: 1 if leap year and month>=3
    movfw   curr_month
    sublw   .9
    skpc
    incf    i2c_adr, F                  ; More than 255 days before month start
    movlw   .3
    subwf   curr_month, W
    bnc     _noCorr                     ; Before march, no leap year correction
    movfw   curr_year                   ; Check for leap year
    andlw   0x03
    skpnz   
    incf    temp2, F
_noCorr
    copyl   HIGH _monthLen, PCLATH      ; Care for jump table
    decf    curr_month, W               ; month 0-11 into W
    call    _monthLen
    addwf   temp2, W    
    movwf   i2c_adr+1
    decf    curr_date, W
    addwf   i2c_adr+1, F
    skpnc
    incf    i2c_adr, F
    ;-----------------------------------
    ; Multiply by 4 (entry size) and add
    ; table start address
    ;-----------------------------------
    clrc
    rlf     i2c_adr+1, F
    rlf     i2c_adr, F
    rlf     i2c_adr+1, F                ; C is 0 here, since hi of i2c_adr was cleared before
    rlf     i2c_adr, W
    addlw   HIGH RISE_TABLE             ; HIGH of table start
    movwf   i2c_adr

    btfsc   clock_mode, 0               ; 0 for sunrise, 1 for sunset
    bsf     i2c_adr+1, 1                ; add offset 2 (bit was 0 before)    
    call    EEStartRead
    call    EEReadNext                  ; read hour from table
    btfss   i2c_temp, 7                 ; is hour invalid?
    goto    _validTime
    call    EEReadLast                  ; Finish IC read anyway
    movlw   .23                         ; "----" for no rise/set at this date 
    call    DisplayWord
    goto    _chooseCursor

_validTime
    ;-----------------------------------
    ; Adjust rise/set time for DST
    ;-----------------------------------
    btfss   flags, FL_DST
    goto    _noAdjust
    call    BCD2Bin                     ; Add 1 hour for DST
    addlw   1
    call    Bin2BCD
_noAdjust
    clrc
    call    ConvertBCD                  ; display it at left side
    call    EEReadLast                  ; Read minute from table
    setc
    call    ConvertBCD                  ; display it at right side
    call    ConvertSegments

_chooseCursor
    movlw   b'11000000'                 ; light 2 left cursors for sunrise
    btfsc   clock_mode, 0
    xorlw   b'11110000'                 ; light 2 right cursors for sunset
    goto    DrawCursor

_monthLen:
    addwf   PCL, F
    retlw   .0                          ; days before Jan, 1
    retlw   .31                         ; days before Feb, 1 etc.
    retlw   .59
    retlw   .90
    retlw   .120
    retlw   .151
    retlw   .181
    retlw   .212
    retlw   .243
    retlw   .273 & 0xff                 ; high byte is handled manually
    retlw   .304 & 0xff
    retlw   .334 & 0xff

;*************************************************************************************************
;*
;* Adjust brightness and speed                                     
;*
;*************************************************************************************************
Brightness
    movfw   pstate                      ; No adjustment in terminal mode
    xorlw   STATE_TERMINAL
    skpnz
    return
    testrbt BTN_VOL_PLUS, _bbr0         ; Increase brightness button pressed?
    movlw   INC_BRIGHT
    addwf   brightness, F               ; Add increment to brightness
    movfw   brightness
    sublw   MAX_BRIGHT                  ; Check for maximum
    skpnc
    goto    _brDone
    copyl   MAX_BRIGHT, brightness      ; Saturation
    goto    _brDone
_bbr0
    testrbt BTN_VOL_MINUS, _bbr1        ; Decrease brightness button pressed?
    movlw   INC_BRIGHT
    subwf   brightness, F               ; Subtract increment
    skpnc                               ; Check for <0
    goto    _brDone
    clrf    brightness                  ; Saturation
_brDone
    copyl   STATE_SHOW_BRIGHT, pstate   ; Display new brightness value for a short time
    copyl   RTC_BRIGHT, FSR             ; Also save new brightness to NV RAM
    movfw   brightness
_adjDone
    call    WriteRtc                    ; Save to NV RAM
_bbr3
    return
_bbr1
    testrbt BTN_PROG_PLUS, _bbr2        ; Increase speed button pressed?
    movlw   INC_SPEED
    addwf   speed, F                    ; Add increment to speed
    movfw   speed
    sublw   MAX_SPEED                   ; Check for maximum
    skpnc
    goto    _spDone
    copyl   MAX_SPEED, speed            ; Saturation
    goto    _spDone
_bbr2
    testrbt BTN_PROG_MINUS, _bbr3       ; Decrease speed button pressed?
    movlw   INC_SPEED
    subwf   speed, F                    ; Subtract increment from speed
    skpnc                               ; Check for <0
    goto    _spDone 
    clrf    speed
_spDone
    copyl   STATE_SHOW_SPEED, pstate    ; Display new speed value for a short time
    copyl   RTC_SPEED, FSR              ; Also save new speed to NV RAM
    movfw   speed
    goto    _adjDone

;*************************************************************************************************
;*
;* Init serial port
;*
;*************************************************************************************************
SerInit
    bsf     STATUS, RP0                 ; Bank 1
    copyl   .12, SPBRG                  ; Baudrate 19200 (async, high speed)
    copyl   b'00100100', TXSTA          ; Async, 8 Bit, enable TX, high speed 
    bcf     STATUS, RP0                 ; Bank 0
    copyl   b'10010000', RCSTA          ; Enable serial, 8 bit, cont. RX
    return  

;*************************************************************************************************
;*
;* Send character in W to serial port
;*
;*************************************************************************************************
SerSend
    btfss   PIR1, TXIF
    goto    SerSend
    movwf   TXREG
    return

;*************************************************************************************************
;*
;* Init real-time clock (enable 1 Hz output)
;* check for RTC NV RAM initialized and init, if not
;* temp2 contains the DST and sunblock flag afterwards
;*
;*************************************************************************************************
RTCInit
    movlw   RTC_SEC                     ; Check clock halt bit (bit 7 of seconds register)
    call    ReadRtc
    movwf   temp
    btfss   temp, 7
    bsf     flags, FL_RTC_OK            ; clock is running

    copyl   RTC_CONTROL, FSR            ; Enable 1 Hz output
    movlw   0x10
    call    WriteRtc
    movlw   RTC_DST                     ; Check DST bit in RTC
    call    ReadRtc
    movwf   temp2
    btfsc   temp2, 0 
    bsf     flags, FL_DST
    movlw   RTC_EYE                     ; Check eye catcher in NV RAM
    call    ReadRtc
    xorlw   EYE_CATCHER
    skpnz
    return 

    ;-----------------------------------
    ; Initialize NV RAM with defaults
    ;-----------------------------------
    movlw   .13                         ; "MSG0"
    call    LoadWord
    copyl   RTC_MESSAGE, FSR
    copyl   .10, temp                   ; 10 default messages to store
    clrf    msg_nr
_saveMsg
    call    SaveRtcWord
    incf    buffer+3, F                 ; MSGn -> MSG(n+1)
    incf    msg_nr, F
    decfsz  temp, F
    goto    _saveMsg
    copyl   RTC_BRIGHT, FSR             ; Store default brightness
    movlw   MAX_BRIGHT
    call    WriteRtc
    copyl   RTC_SPEED, FSR              ; Store default speed
    movlw   .10
    call    WriteRtc
    copyl   RTC_MSG_NR, FSR             ; Store default message number
    movlw   0
    call    WriteRtc
    copyl   RTC_PSTATE, FSR             ; Store default pstate
    movlw   STATE_WORDS
    call    WriteRtc
    copyl   RTC_WORD_MODE, FSR          ; Store default word_mode
    movlw   1<<WM_DICT                  
    call    WriteRtc
    copyl   RTC_CLOCK_MODE, FSR         ; Store default clock_mode
    movlw   0
    call    WriteRtc
    copyl   RTC_EYE, FSR                ; Mark NV RAM as initialized
    movlw   EYE_CATCHER
    call    WriteRtc
    return

;*************************************************************************************************
;*
;* Load parameters from NV RAM
;*
;*************************************************************************************************
LoadNVParms
    movlw   RTC_PSTATE                  ; Load pstate
    call    ReadRtc
    movwf   pstate
    movlw   RTC_BRIGHT                  ; Load brightness
    call    ReadRtc
    movwf   brightness
    movlw   RTC_SPEED                   ; Load speed
    call    ReadRtc
    movwf   speed
    movlw   RTC_MSG_NR                  ; Load message number
    call    ReadRtc
    movwf   msg_nr
    movlw   RTC_WORD_MODE               ; Load word_mode
    call    ReadRtc
    movwf   word_mode
    movlw   RTC_CLOCK_MODE              ; Load clock_mode
    call    ReadRtc
    movwf   clock_mode
    return

;*************************************************************************************************
;*
;* Delay W 1/10 seconds (0.1 - 25.5s)
;*
;*************************************************************************************************
delayTenth
    movwf   delay3
_d3
    movlw   .100
    call    delayMs
    decfsz  delay3, F
    goto    _d3
    return

;*************************************************************************************************
;*
;* Delay W milliseconds (1 - 255 ms)
;*
;*************************************************************************************************
delayMs
    movwf   delay1 
_d1
    copyl   .110, delay2                ; 1 ms
_d2
    nop 
    nop
    nop
    nop
    nop
    nop
    decfsz  delay2, F
    goto    _d2
    decfsz  delay1, F
    goto    _d1
    return

;*************************************************************************************************
;*
;* Initialize wframe counter from speed setting -- could be fancier :-)
;*
;*************************************************************************************************
InitWFrame
    copy    speed, wframe
    incf    wframe, F
    return

;*************************************************************************************************
;*
;* Convert buffer to lower/title case depending on word_mode
;*
;*************************************************************************************************
HandleCase:
    btfss   word_mode, WM_FIRST_LOWER
    goto    _hcRest
    btfsc   buffer, ALPHA_BIT           ; Only when alphabetic... 
    bsf     buffer, CASE_BIT            ; ...make it lower case
_hcRest
    btfss   word_mode, WM_REST_LOWER
    return
    btfsc   buffer+1, ALPHA_BIT         ; Only when alphabetic...
    bsf     buffer+1, CASE_BIT          ; ...make it lower case
    btfsc   buffer+2, ALPHA_BIT
    bsf     buffer+2, CASE_BIT
    btfsc   buffer+3, ALPHA_BIT
    bsf     buffer+3, CASE_BIT
    return

;*************************************************************************************************
;*
;* Load/display word with index W from internal EEPROM
;*
;*************************************************************************************************
LoadWord    
    movwf   FSR                         ; Multiply word index by 4
    clrc
    rlf     FSR, F
    clrc
    rlf     FSR, F
    call    ReadIntEEPROM
    movwf   buffer
    incf    FSR, F
    call    ReadIntEEPROM
    movwf   buffer+1
    incf    FSR, F
    call    ReadIntEEPROM
    movwf   buffer+2
    incf    FSR, F
    call    ReadIntEEPROM
    movwf   buffer+3
    goto    HandleCase                  ; tail call

DisplayWord
    call    LoadWord
    ; Fall thru!

;*************************************************************************************************
;*
;* Convert char buffer to segments
;*
;*************************************************************************************************
ConvertSegments
    copyl   HIGH convert_hi, PCLATH     ; Implies segupper
    movfw   buffer
    call    convert_hi
    movwf   segments
    movfw   buffer+1
    call    convert_hi
    movwf   segments+2
    movfw   buffer+2
    call    convert_hi
    movwf   segments+4
    movfw   buffer+3
    call    convert_hi
    movwf   segments+6

    copyl   HIGH convert_lo, PCLATH     ; Implies segupper
    movfw   buffer
    call    convert_lo
    movwf   segments+1
    movfw   buffer+1
    call    convert_lo
    movwf   segments+3
    movfw   buffer+2
    call    convert_lo
    movwf   segments+5
    movfw   buffer+3
    call    convert_lo
    movwf   segments+7
    seglower
    return 

;*************************************************************************************************
;*
;* Convert char buffer to hidden segments
;*
;*************************************************************************************************
ConvertHiddenSegments
    copyl   HIGH convert_hi, PCLATH     ; Implies segupper
    movfw   buffer
    call    convert_hi
    movwf   hsegments
    movfw   buffer+1
    call    convert_hi
    movwf   hsegments+2
    movfw   buffer+2
    call    convert_hi
    movwf   hsegments+4
    movfw   buffer+3
    call    convert_hi
    movwf   hsegments+6

    copyl   HIGH convert_lo, PCLATH     ; Implies segupper
    movfw   buffer
    call    convert_lo
    movwf   hsegments+1
    movfw   buffer+1
    call    convert_lo
    movwf   hsegments+3
    movfw   buffer+2
    call    convert_lo
    movwf   hsegments+5
    movfw   buffer+3
    call    convert_lo
    movwf   hsegments+7
    seglower
    return 

;*************************************************************************************************
;*
;* Blank all segments
;*
;*************************************************************************************************
ClearSegments
    clrf    segments+0
    clrf    segments+1
    clrf    segments+2
    clrf    segments+3
    clrf    segments+4
    clrf    segments+5
    clrf    segments+6
    clrf    segments+7
    return

;*************************************************************************************************
;*
;* Copy animation pattern from EEPROM, W bytes starting at FSR
;*
;*************************************************************************************************
CopyAnim
    movwf   bitcount
    copy    FSR, temp
    copyl   seg_anim, temp1
_rcopy
    copy    temp, FSR
    call    ReadIntEEPROM
    movwf   temp2
    copy    temp1, FSR
    copy    temp2, INDF
    incf    temp, F
    incf    temp1, F
    decfsz  bitcount, F
    goto    _rcopy
    return

;*************************************************************************************************
;*
;* Convert BCD in W to buffer (left 2 chars if C=0, right 2 chars if C=1)
;*
;*************************************************************************************************
ConvertBCD
    movwf   temp
    bc      _toRight
    andlw   0x0f
    addlw   '0'
    movwf   buffer+1
    swapf   temp, W
    andlw   0x0f
    addlw   '0'
    movwf   buffer
    return
_toRight
    andlw   0x0f
    addlw   '0'
    movwf   buffer+3
    swapf   temp, W
    andlw   0x0f
    addlw   '0'
    movwf   buffer+2
    return

;----------------------------------------------------------
; Test if num >= num2: C=1
;----------------------------------------------------------
Compare16
    movfw   num2
    subwf   num, W
    skpz
    return
    movfw   num2+1
    subwf   num+1, W
    return


;*************************************************************************************************
;*************************************************************************************************
;*
;* Serial EEPROM access 
;*
;*************************************************************************************************
;*************************************************************************************************
EEStartRead                             ; Set EEPROM read address to i2c_adr
    call    I2C_START
    movlw   EEPWrite                    ; SIC!
    call    I2C_SEND
    call    I2C_DEV_ACK
    movfw   i2c_adr
    call    I2C_SEND
    call    I2C_DEV_ACK
    movfw   i2c_adr+1
    call    I2C_SEND
    call    I2C_DEV_ACK

    call    I2C_START
    movlw   EEPRead 
    call    I2C_SEND
    call    I2C_DEV_ACK
    return

EEReadNext                              ; Read next byte sequentially
    call    I2C_RECEIVE
    call    I2C_M_ACK
    return

EEReadLast                              ; Read next byte and terminate reading
    call    I2C_RECEIVE
    call    I2C_M_NACK
    call    I2C_STOP
    return 

;*************************************************************************************************
;*
;* Write 64 byte from eebuf to EEPROM
;*
;*************************************************************************************************
WriteEEBuf
    bcf     INTCON, GIE
    movwf   i2c_byte
    call    I2C_START
    movlw   EEPWrite
    call    I2C_SEND
    call    I2C_DEV_ACK
    movfw   i2c_adr
    call    I2C_SEND
    call    I2C_DEV_ACK
    movfw   i2c_adr+1
    call    I2C_SEND
    call    I2C_DEV_ACK
    copyl   .64, eecount
    copyl   eebuf, FSR
_wloop
    movfw   INDF
    call    I2C_SEND
    call    I2C_DEV_ACK
    incf    FSR, F
    decfsz  eecount, F
    goto    _wloop
    call    I2C_STOP
    bsf     INTCON, GIE
    return 

;*************************************************************************************************
;*
;* Write eecount bytes from eebuf to RTC, starting at address in FSR
;*
;*************************************************************************************************
WriteClock
    bcf     INTCON, GIE
    movwf   i2c_byte
    call    I2C_START
    movlw   RTCWrite
    call    I2C_SEND
    call    I2C_DEV_ACK
    movfw   FSR                         ; Start address of RTC/NVRAM
    call    I2C_SEND
    call    I2C_DEV_ACK
    copyl   eebuf, FSR
_wloop1
    movfw   INDF
    call    I2C_SEND
    call    I2C_DEV_ACK
    incf    FSR, F
    decfsz  eecount, F
    goto    _wloop1
    call    I2C_STOP
    bsf     INTCON, GIE
    return 

;*************************************************************************************************
;*
;* Read RTC register number in W into W
;*
;*************************************************************************************************
ReadRtc
    movwf   rtc_temp
    call    I2C_START
    movlw   RTCWrite
    call    I2C_SEND
    call    I2C_DEV_ACK
    movfw   rtc_temp
    call    I2C_SEND
    call    I2C_DEV_ACK
    call    I2C_START
    movlw   RTCRead 
    call    I2C_SEND
    call    I2C_DEV_ACK
    call    I2C_RECEIVE
    movwf   rtc_temp
    call    I2C_M_NACK
    call    I2C_STOP
    movfw   rtc_temp
    return 

;*************************************************************************************************
;*
;* Write W to RTC register FSR
;*
;*************************************************************************************************
WriteRtc
    movwf   rtc_temp
    call    I2C_START
    movlw   RTCWrite
    call    I2C_SEND
    call    I2C_DEV_ACK
    movfw   FSR
    call    I2C_SEND
    call    I2C_DEV_ACK
    movfw   rtc_temp
    call    I2C_SEND
    call    I2C_DEV_ACK
    call    I2C_STOP
    return 

;*************************************************************************************************
;*
;* Compute start of word in NV RAM (ignoring blinking bit)
;*
;*************************************************************************************************
ComputeWordAddress
    copy    msg_nr, FSR         
    clrc
    rlf     FSR, F                      ; Blinking bit rotated out
    clrc                                ; don't let it be rotated back in!
    rlf     FSR, W
    addlw   RTC_MESSAGE
    movwf   FSR
    return

;*************************************************************************************************
;*
;* Load word with index msg_nr from NV RAM into buffer
;*
;*************************************************************************************************
LoadRtcWord
    call    ComputeWordAddress
    movfw   FSR
    call    ReadRtc
    movwf   buffer
    incf    FSR, f
    movfw   FSR
    call    ReadRtc
    movwf   buffer+1
    incf    FSR, f
    movfw   FSR
    call    ReadRtc
    movwf   buffer+2
    incf    FSR, f
    movfw   FSR
    call    ReadRtc
    movwf   buffer+3
    incf    FSR, f
    return

;*************************************************************************************************
;*
;* Write word in buffer to NV RAM at index msg_nr
;*
;*************************************************************************************************
SaveRtcWord
    call    ComputeWordAddress
    movfw   buffer
    call    WriteRtc
    incf    FSR, f
    movfw   buffer+1
    call    WriteRtc
    incf    FSR, f
    movfw   buffer+2
    call    WriteRtc
    incf    FSR, f
    movfw   buffer+3
    call    WriteRtc
    incf    FSR, f
    return  

;*************************************************************************************************
;*************************************************************************************************
;*
;* IC bus access library
;*
;*************************************************************************************************
;*************************************************************************************************
I2C_DELAY                               ; Needed by the library
    nop                                 ; for proper timing of
    nop                                 ; I2C bus.  
    nop
    nop
    return
    
    #include "I2C_M_3pins.inc"

;*************************************************************************************************
;*************************************************************************************************
;*
;* Random number generation by a Linear Shift Register
;* 47 bit shift register with additional tap at bit 42
;* Result in RAND and W
;*
;*************************************************************************************************
;*************************************************************************************************
Random
    copyl   .8, bitcount                ; Need to generate 8 random bits
NextRbit
    rrf     LSR, F                      ; High byte, no matter what's rotated in
    rrf     LSR+1, F            
    rrf     LSR+2, F            
    rrf     LSR+3, F            
    rrf     LSR+4, F            
    rrf     LSR+5, F                    ; Low byte, C is output
    rlf     RAND, F                     ; One bit of result
    bcf     LSR, 6                      ; Clear highest bit of LSR
    movfw   RAND                        ; Need only lowest bit (LBit)
    btfsc   LSR+5, 4                    ; Bit 42 of unshifted register
    comf    RAND, W                     ; was set, so invert LBit
    movwf   temp
    btfsc   temp, 0                     ; LBit clear, so don't set highest bit
    bsf     LSR, 6                      ; Set highest bit of LSR 
    decfsz  bitcount, F
    goto    NextRbit
    movfw   RAND
    return

;*************************************************************************************************
;*
;* Select a random number in [0..W[
;*
;*************************************************************************************************
RandomRange
    movwf   temp1
_nextTry
    copy    temp1, temp
    clrf    RAND
    copyl   .9, bitcount
_nextrrbit                              ; Determine highest bit set
    decf    bitcount, F                 ; and generate only remaining bits
    rlf     temp, F
    bnc     _nextrrbit
    call    NextRbit
    movfw   temp1
    subwf   RAND, W
    bc      _nextTry                    ; Out of range? Try again.
    movfw   RAND
    return

;*************************************************************************************************
;*
;* Init LSFR from RTC and waste some randoms to shuffle LSR
;*
;*************************************************************************************************
RandomInit
    movlw   0
    call    ReadRtc
    movwf   LSR+5
    movlw   1
    call    ReadRtc
    movwf   LSR+4
    movlw   2
    call    ReadRtc
    movwf   LSR+3
    movlw   3
    call    ReadRtc
    movwf   LSR+2
    movlw   4
    call    ReadRtc
    movwf   LSR+1
    movlw   5
    call    ReadRtc
    movwf   LSR
    ;-----------------------------------
    ; Skip some randoms
    ;-----------------------------------
    movlw   .213                        ; What's this number? :-)
    movwf   temp1
_wasteRand  
    call    Random
    decfsz  temp1, F
    goto    _wasteRand
    return

;*************************************************************************************************
;*
;* Init word generation
;*
;*************************************************************************************************
WordInit
    bcf     flags, FL_ROM_OK            ; Assume invalid EEPROM
    clrf    i2c_adr                     ; Read first 3 bytes of EEPROM:
    clrf    i2c_adr+1
    call    EEStartRead
    call    EEReadNext                  ; Check eyecatcher (byte 1)
    xorlw   EYE_CATCHER
    skpnz
    bsf     flags, FL_ROM_OK            ; EEPROM (probably) OK
    call    EEReadNext
    movwf   num_words                   ; byte 2 is HI num_words
    call    EEReadLast
    movwf   num_words+1                 ; byte 3 is LO num_words
    clrf    chain_ptr
    clrf    chain_ptr+1
    copyl   1, wframe

    ;-----------------------------------
    ; Fill seg_nr array with 0..13
    ;-----------------------------------
    copyl   .13, temp
    copyl   seg_nr, FSR
_fillSeg
    copy    temp, INDF
    incf    FSR, F
    decfsz  temp, F
    goto    _fillSeg
    clrf    INDF
    return

;*************************************************************************************************
;*
;* Select a random word from dictionary in EEPROM
;*
;*************************************************************************************************
SelectWord
    btfsc   flags, FL_ROM_OK
    goto    _romOK
    movlw   .0
    goto    DisplayWord                 ; "ROM?"

_romOK
    copy2   num_words, num2

_newTrial
    incf    num_words, W                ; Hi byte+1 
    call    RandomRange
    movwf   num
    call    Random
    movwf   num+1
    call    Compare16
    bc      _newTrial                   ; now 0<= r=num < num_words,
    ;-----------------------------------
    ; multiply by 6 (4 letters, 2 byte pointer)
    ;-----------------------------------
    rlf     num+1, W                    ; C=0 asserted by branch before
    movwf   num2+1
    rlf     num, W
    movwf   num2                        ; num2 = 2*r
    movfw   num2+1
    addwf   num+1, F
    skpnc
    incf    num, F
    movfw   num2
    addwf   num, F                      ; num = 3*r
    clrc    
    rlf     num+1, F            
    rlf     num, F                      ; num = 6*r

    movlw   7                           ; Memory before words start in EEPROM
    addwf   num+1, F
    skpnc
    incf    num, F

    copy2   num, i2c_adr
    call    FetchWord   
    copyl   1, num_trans                ; Allow chain break immediately
    copy2   chain_ptr, last_chain_ptr
    return

;----------------------------------------------------------
; Read word and chain_ptr from EEPROM at i2c_adr
;----------------------------------------------------------
FetchWord
    clrf    diff_char
    call    EEStartRead                 ; Fetch pointer to transition table
    call    EEReadNext                  
    movwf   chain_ptr
    call    EEReadNext
    movwf   chain_ptr+1
    call    EEReadNext                  ; and the word itself
    savec   0
    call    EEReadNext
    savec   1
    call    EEReadNext
    savec   2
    call    EEReadLast
    savec   3
    return

;----------------------------------------------------------
; Read chain_ptr from EEPROM at i2c_adr
;----------------------------------------------------------
FetchWordPtr
    call    EEStartRead                 ; Fetch pointer to transition table
    call    EEReadNext                  
    movwf   chain_ptr
    call    EEReadLast
    movwf   chain_ptr+1
    return

;*************************************************************************************************
;*
;* Select a random word chained from the current one
;*
;*************************************************************************************************
SelectChainedWord
    ;-----------------------------------
    ; check for null chain_ptr
    ;-----------------------------------
    movf    chain_ptr, F                ; =0?
    bnz     _validPtr
    movf    chain_ptr+1, F              ; =0?
    bnz     _validPtr
    goto    SelectWord                  ; null ptr, so select random word
                                        ; from dict first
_validPtr
    copy2   chain_ptr, chain_ptr_bak
    copy2   chain_ptr, i2c_adr
_retryRand
    call    EEStartRead         
    call    EEReadLast                  ; Read number of potential successors
    movwf   num_trans
    andlw   0x7f                        ; mask actual length
    bz      _chainEnd                   ; 0 means isolated word
    call    RandomRange
    incf    RAND, F                     ; +1 (overread length)
    addwf   RAND, W                     ; 1+2*rand(0..length)
    addwf   i2c_adr+1, F
    skpnc
    incf    i2c_adr, F
    call    EEStartRead                 ; Read ptr to next word 
    call    EEReadNext
    movwf   i2c_adr
    call    EEReadLast      
    movwf   i2c_adr+1
    copy2   i2c_adr, word_ptr           ; Store to retrieve word itself later
    call    FetchWordPtr    
    movfw   last2_chain_ptr             ; Is this the same word 2 steps ago?
    xorwf   chain_ptr, W
    bnz     _noRepeat
    movfw   last2_chain_ptr+1
    xorwf   chain_ptr+1, W
    bnz     _noRepeat
    movfw   num_trans                   ; Yes: Was this the only choice?
    andlw   0x7f
    xorlw   1
    bz      _chainEnd                   ; Yes: end the chain here
    copy2   chain_ptr_bak, i2c_adr      ; No: restore original read position
    goto    _retryRand                  ; and try another chained word

_chainEnd
    ;-----------------------------------
    ; Last word repeated, break chain
    ;-----------------------------------
    clrf    chain_ptr
    clrf    chain_ptr+1
    copyl   0xff, diff_char
    movlw   .12                         ; "****"
    goto    LoadWord     

_noRepeat           
    ;-----------------------------------
    ; Check for "small cycle" bit and break randomly
    ;-----------------------------------
    btfss   num_trans, 7
    goto    _regularWord
    movlw   .8
    call    RandomRange
    bz     _chainEnd

    ;-----------------------------------
    ; Read word itself
    ;-----------------------------------
_regularWord
    copy2   word_ptr, i2c_adr
    call    FetchWord   

    ;-----------------------------------
    ; Shift 2 words stored formerly
    ;-----------------------------------
    copy2   last_chain_ptr, last2_chain_ptr 
    copy2   chain_ptr, last_chain_ptr 
    return

;----------------------------------------------------------
; Fetch ROM tag to buffer
;----------------------------------------------------------
FetchROMTag
    btfsc   flags, FL_ROM_OK
    goto    _romOK1
    movlw   .0
    goto    DisplayWord
_romOK1
    clrf    i2c_adr                     ; Adress of tag
    copyl   3, i2c_adr+1 
    call    EEStartRead                 
    call    EEReadNext                  
    movwf   buffer
    call    EEReadNext
    movwf   buffer+1
    call    EEReadNext
    movwf   buffer+2
    call    EEReadLast
    movwf   buffer+3
    return

;*************************************************************************************************
;*
;* Read PIC internal EEPROM cell whose address is in FSR into W
;*
;*************************************************************************************************
ReadIntEEPROM
    bsf     STATUS, RP0                 ; Register page 1
    copy    FSR, EEADR                  ; Set EEPROM address
    bsf     EECON1, RD
    movf    EEDATA, W
    bcf     STATUS, RP0                 ; Restore register page
    return

;*************************************************************************************************
;*
;* Create byte with single bit (nr in W) set
;*
;*************************************************************************************************
BitShift
    addwf   PCL, F
    retlw   b'10000000'
    retlw   b'01000000'
    retlw   b'00100000'
    retlw   b'00010000'
    retlw   b'00001000'
    retlw   b'00000100'
    retlw   b'00000010'
    retlw   b'00000001'

;*************************************************************************************************
;*
;* Convert BCD in W to binary in W
;*
;*************************************************************************************************
BCD2Bin
    movwf   temp1
    andlw   0x0f
    btfsc   temp1, 4
    addlw   .10
    btfsc   temp1, 5
    addlw   .20
    btfsc   temp1, 6
    addlw   .40
    btfsc   temp1, 7
    addlw   .80
    return

;*************************************************************************************************
;*
;* Convert binary in W to BCD in W (must be <= 99)
;*
;* (but look at http://www.piclist.com/techref/method/math/radixs.htm  
;* for an alternative)
;*
;*************************************************************************************************
Bin2BCD
    movwf   temp1
    clrf    temp2
_l10
    movlw   .10
    subwf   temp1, W
    bnc     _done
    movwf   temp1
    incf    temp2, F
    goto    _l10
_done
    swapf   temp2, W
    iorwf   temp1, W
    return


;*************************************************************************************************
;*
;* 70ies-style word generation: Each tube gets a (different)
;* randomly selected letter.
;*
;*************************************************************************************************
Make70iesWord
    copyl   HIGH _randLetter1, PCLATH
    clrf    diff_char
    call    _randLetter1
    savec   0
    call    _randLetter2
    savec   1
    call    _randLetter3
    savec   2
    call    _randLetter4
    savec   3
    return

_randLetter1:
    movlw   .10
    call    RandomRange
    addwf   PCL, F
    dt      "BCDFHLPRST"

_randLetter2:
    movlw   .8
    call    RandomRange
    addwf   PCL, F
    dt      "AEHILORU"

_randLetter3:
    movlw   .10
    call    RandomRange
    addwf   PCL, F
    dt      "ACEILNORST"

_randLetter4:
    movlw   .10
    call    RandomRange
    addwf   PCL, F
    dt      "DEKLNPRSTY"

;*************************************************************************************************
;*************************************************************************************************
;*************************************************************************************************
;*
;* UPPER 2k PAGE --- UPPER 2k PAGE --- UPPER 2k PAGE --- UPPER 2k PAGE --- UPPER 2k PAGE --- UPPER
;*
;*************************************************************************************************
;*************************************************************************************************
;*************************************************************************************************

    org 0x800

;*************************************************************************************************
;*************************************************************************************************
;*
;* Interrupt handler (continued)
;*
;*************************************************************************************************
;*************************************************************************************************
IRHandler
    btfsc   INTCON, RBIF                ; Interrupt source port B change?
    goto    _RBChange
    btfsc   PIR1, RCIF                  ; Interrupt source serial receive?
    goto    _SerReceive
    btfsc   PIR1, TMR2IF                ; Interrupt source Timer2 overflow?
    goto    _Timer2Over
    btfsc   INTCON, T0IF                ; Interrupt source Timer0 overflow?
    goto    _Timer0Over
    goto    _IRHandled

_Timer2Over
    ;-----------------------------------
    ; Timer2 overflow: Update tubes
    ;-----------------------------------
    movfw   FSR                         ; FSR is used by UpdateTube, so save it
    movwf   FSR_temp
    call    UpdateTubes
    movfw   FSR_temp
    movwf   FSR
    bcf     PIR1, TMR2IF                ; re-enable timer2 interrupts
    goto    _IRHandled

_SerReceive
    ;-----------------------------------
    ; RS232 receive
    ;-----------------------------------
    btfsc   RCSTA, FERR                 ; Ignore receive errors
    goto    _IRHandled
    btfsc   RCSTA, OERR
    goto    _IRHandled
    movfw   pstate                      ; check what to do, depending on pstate
    xorlw   STATE_DOWNLOAD_ROM
    bz      _addBuffer
    movfw   pstate
    xorlw   STATE_DOWNLOAD_TIME
    bz      _addBuffer
    movfw   pstate
    xorlw   STATE_DOWNLOAD_MSG
    bz      _addBuffer
    movfw   pstate
    xorlw   STATE_DOWNLOAD_LEN
    bz      _recvLen
    movfw   pstate
    xorlw   STATE_TERMINAL
    bz      _shiftDisplay
    movfw   RCREG                       ; Char not usable now, read and ignore
    goto    _IRHandled

_shiftDisplay                           ; Add char to display buffer, shift left
    copy    buffer+1, buffer
    copy    buffer+2, buffer+1
    copy    buffer+3, buffer+2
    movfw   RCREG
    andlw   0x7f                        ; Eighth bit would overflow font table
    movwf   buffer+3
    goto    _IRHandled

_recvLen                                ; We received a length byte
    movfw   RCREG
    btfss   recv_count, 1       
    movwf   num_blocks+1
    btfss   recv_count, 0       
    movwf   num_blocks
    decfsz  recv_count, F
    goto    _IRHandled
    copyl   STATE_DOWNLOAD_ROM, pstate  ; 2 length bytes received, now download
    movlw   ACK                         ; Send ack to PC
    seglower
    call    SerSend
    segupper
    clrf    i2c_adr                     ; Set EEPROM write address          
    clrf    i2c_adr+1
    copyl   eebuf, eebufp               ; Prepare to receive 64 byte block
    copyl   .64, recv_count
    goto    _IRHandled

_addBuffer                              ; Add received char to buffer
    copy    FSR, FSR_temp               ; Save FSR register
    copy    eebufp, FSR
    copy    RCREG, INDF                 ; put received char into buffer
    incf    eebufp, F
    decf    recv_count, F
    skpnz
    incf    pstate, F                   ; full block received: download state -> saving state
    copy    FSR_temp, FSR               ; restore FSR register
    goto    _IRHandled

_RBChange
    ;-----------------------------------
    ; IR signal change: Measure pulse
    ; length and add bit to IR receiver
    ;-----------------------------------
    movfw   PORTB                       ; MUST read to allow resetting interrupt flag!
    btfss   rc_state, 0
    goto    _firstEdge                  ; rc_state==READY
    btfss   rc_state, 1
    goto    _measureWidth               ; Subsequent edge
    clrf    rc_state
    goto    _rbHandled                  ; rc_state==WAIT: edge too early: Reset timer

_firstEdge
    btfsc   RC5_INPUT                   ; First half, sample RC pin
    goto    _wrongFirst 
    clrf    rc_bit                      ; Reset receiver variables
    clrf    rc_buf_hi
    clrf    rc_buf_lo
    incf    rc_bit, F                   ; Start bit, assume short pulse
    bsf     rc_state, 0                 ; rc_state=RECEIVING, we are receiving now
    goto    _received1

_wrongFirst
    bsf     rc_state, 0                 ; rc_state=WAIT
    bsf     rc_state, 1                 ; rc_state=WAIT
    goto    _rbHandled

_measureWidth
    movfw   TMR0
    sublw   .10                         ; very short pulse, ignore spike
    bc      _rbHandled
    movfw   TMR0            
    sublw   .160                        ; short: 889 s, long 1778 s, cut at 160 * 8s 
    skpc
    incf    rc_bit, F                   ; Long pulse, crossing an edge, so correct edge count
    btfss   rc_bit, 0                   ; Odd edge counter?
    goto    _checkComplete              ; -> second half of bit
    btfss   RC5_INPUT                   ; First half, sample RC pin
    goto    _received1
    clrc                                ; High -> 0 received
    goto    _rotateBit
_received1
    setc                                ; Low -> 1 received
_rotateBit
    rlf     rc_buf_lo, F                ; Rotate IR bit into receive buffer
    rlf     rc_buf_hi, F    

_checkComplete
    movfw   rc_bit
    sublw   .27                         ; 14 bits (27 edges) received, 
    bnz     _recvNext

    ;-----------------------------------
    ; Complete code received, we 
    ; currently ignore the group code
    ;-----------------------------------
    movfw   rc_buf_lo
    andlw   0x3f                        ; retain low 6 bits
    movwf   rc_cmd_rep
    btfss   rc_buf_hi, 4                ; 7th cmd bit (negated) in start sequence
    bsf     rc_cmd_rep, 6
    rlf     rc_toggle, F                ; Last toggle now in bit pos 1
    bcf     rc_toggle, 0
    btfsc   rc_buf_hi, 3                ; toggle bit set?
    bsf     rc_toggle, 0                ; store it
    incf    rc_toggle, W                ; 0-3 -> 1-4; bit 1 == 1 <=> was 1 or 2 previously
    movwf   rc_bit                      ; variable re-used
    movfw   rc_cmd_rep
    btfsc   rc_bit, 1                   ; Toggle bit different to previous
    movwf   rc_cmd                      ; so set non-repeated command, too
    bsf     rc_state, 1                 ; Now WAITING (bit 0 already set)
    goto    _rbHandled
_recvNext
    incf    rc_bit, F                   ; Next (half-) bit
_rbHandled
    bcf     INTCON, RBIF 
    clrf    TMR0                        ; restart timer
    goto    _IRHandled

_Timer0Over
    ;-----------------------------------
    ; Timer0 overflow: modify IR
    ; receiver state
    ;-----------------------------------
    btfss   rc_state, 0
    goto    _t0Handled                  ; rc_state==0: ignore interrupts
    btfsc   rc_state, 1                 ; Are be waiting or receiving?
    goto    _timerWait                  ; rc_state==3: finish waiting
    bsf     rc_state, 1                 ; rc_state==2: now wait: rc_state=3
    goto    _t0Handled
_timerWait
    clrf    rc_state                    ; Set state to READY
_t0Handled
    bcf     INTCON, T0IF                ; re-enable timer0 interrupts
    ; fall thru

_IRHandled
    ;-----------------------------------
    ; Interrupt handling finished
    ;-----------------------------------
    copy    PCLATH_temp, PCLATH         ; Standard epilog:
    swapf   STATUS_temp, W              ; restore W, STATUS and PCLATH
    movwf   STATUS
    swapf   W_temp, F
    swapf   W_temp, W
    retfie

;*************************************************************************************************
;*
;* Update tube display: the next tube is blanked or powered 
;* with varying duty cycle 
;*
;*************************************************************************************************
UpdateTubes
    incf    tube_phase, F               ; Toggle on/off phase
    btfss   tube_phase, 0
    goto    _blankAnode
    clrc
    rrf     anodes, F                   ; Select next anode
    btfss   anodes, 3                   ; Right-most tube exceeded?
    goto    _cont1
    copyl   b'10000000', anodes         ; Start left again
    copyl   segments, segp              ; and reinit pointer to segments
_cont1
    clrf    pattern                     ; Switch off anodes
    outPat  pattern
    bsf     TUBE_ANODES                 ; Pulse anode strobe
    bcf     TUBE_ANODES
    copy    segp, FSR                   ; Read 16 segment bits and advance ptr
    copy    INDF,tube_data
    incf    FSR, F
    copy    INDF, tube_data+1
    incf    FSR, W
    movwf   segp

    ;-----------------------------------
    ; Output 16 bit cathode pattern
    ;-----------------------------------
    outPat  tube_data
    outPat  tube_data+1
    bsf     TUBE_CATHODES               ; Pulse cathode strobe
    bcf     TUBE_CATHODES               ; 

    ;-----------------------------------
    ; Output 8 bit anode pattern
    ;-----------------------------------
    copy    anodes, pattern
    outPat  pattern
    bsf     TUBE_ANODES                 ; Pulse anode strobe
    bcf     TUBE_ANODES
    goto    _setTimer2

_blankAnode
    clrf    pattern                     ; Clear anodes
    outPat  pattern
    bsf     TUBE_ANODES                 ; Pulse anode strobe
    bcf     TUBE_ANODES

_setTimer2
    ;-----------------------------------
    ; Determine time until next interrupt
    ;-----------------------------------
    copyl   HIGH _tubeOnTimer, PCLATH
    movfw   brightness
    btfss   tube_phase, 0
    call    _tubeOnTimer
    btfsc   tube_phase, 0
    call    _tubeOffTimer
    bsf     STATUS, RP0
    movwf   PR2                         ; Set target for timer 2
    bcf     STATUS, RP0
    return

;----------------------------------------------------------
; Timer delays for each brightness level. There's no systematic approach here,
; just trial and error. Some harmless looking on/off times yield wild flicker.  
;----------------------------------------------------------
_tubeOnTimer
    addwf   PCL, F
    dt      .3, .3, .4, .4,.20,.40, .6

_tubeOffTimer
    addwf   PCL, F
    dt      .7,.10,.12,.20,.20, .0, .0

;*************************************************************************************************
;*************************************************************************************************
;*
;* Interprete rc command to edit current tube                
;*
;*************************************************************************************************
;*************************************************************************************************
EditTube
    movfw   edit_tube                   ; Load address of current char into FSR
    addlw   buffer
    movwf   FSR

    testrbt BTN_CURS_UP, _et1           ; Increase char of current tube
    incf    INDF, W
    andlw   0x7f
    movwf   INDF
    goto    EditDone

_et1
    testrbt BTN_CURS_DOWN, _et2         ; Decrease char of current tube
    decf    INDF, W
    andlw   0x7f
    movwf   INDF
    goto    EditDone

_et2
    testbut BTN_DIGIT_1, _et3
    change  ' ', '1'
    copyl   ' ', INDF
    goto    EditDone

_et3
    testbut BTN_DIGIT_2, _et4
  if GERMAN
    change  'a', '{'                    ; 
    change  '{', 'b'
  else
    change  'a', 'b'
  endif
    change  'b', 'c'
    change  'c', '2'
  if GERMAN
    change  'A', '['                    ; 
    change  '[', 'B'
  else
    change  'A', 'B'
  endif
    change  'B', 'C'
    change  'C', '2'
    copyl   'A', INDF
    goto    EditDone

_et4
    testbut BTN_DIGIT_3, _et5
    change  'd', 'e'
    change  'e', 'f'
    change  'f', '2'
    change  'D', 'E'
    change  'E', 'F'
    change  'F', '3'
    change  '3', '\x03'                 ; big Delta
    change  '\x03', '\x04'              ; small delta
    change  '\x04', '\x05'              ; Euro
    copyl   'D', INDF
    goto    EditDone

_et5
    testbut BTN_DIGIT_4, _et6
    change  'g', 'h'
    change  'h', 'i'
    change  'i', '4'
    change  'G', 'H'
    change  'H', 'I'
    change  'I', '4'
    change  '4', '\x07'                 ; big Gamma
    copyl   'G', INDF
    goto    EditDone

_et6
    testbut BTN_DIGIT_5, _et7
    change  'j', 'k'
    change  'k', 'l'
    change  'l', '5'
    change  'J', 'K'
    change  'K', 'L'
    change  'L', '5'
    change  '5', '\x0c'                 ; small lambda
    copyl   'J', INDF
    goto    EditDone

_et7
    testbut BTN_DIGIT_6, _et8
    change  'm', 'n'
    change  'n', 'o'
  if GERMAN
    change  'o', '|'                    ; 
    change  '|', '6'
  else
    change  'o', '6'
  endif
    change  'M', 'N'
    change  'N', 'O'
  if GERMAN
    change  'O', '\\'                   ; 
    change  '\\', '6'
  else
    change  'O', '6'
  endif
    change  '6', '\x0d'                 ; small my
    change  '\x0d', '\x0f'              ; big Omega
    copyl   'M', INDF
    goto    EditDone

_et8
    testbut BTN_DIGIT_7, _et9
    change  'p', 'q'
    change  'q', 'r'
    change  'r', 's'
    change  's', '7'
    change  'P', 'Q'
    change  'Q', 'R'
    change  'R', 'S'
    change  'S', '7'
    change  '7', '\x02'                 ; Sharp-S ()
    change  '\x02', '\x10'              ; small pi
    change  '\x10', '\x11'              ; big Pi
    change  '\x11', '\x12'              ; Root sign
    change  '\x12', '\x13'              ; big Sigma
    copyl   'P', INDF
    goto    EditDone

_et9
    testbut BTN_DIGIT_8, _et10
    change  't', 'u'
  if GERMAN
    change  'u', '}'                    ; 
    change  '}', 'v'
  else
    change  'u', 'v'
  endif
    change  'v', '8'
    change  'T', 'U'
  if GERMAN
    change  'U', ']'                    ; 
    change  ']', 'V'
  else
    change  'U', 'V'
  endif
    change  'V', '8'
    copyl   'T', INDF
    goto    EditDone

_et10
    testbut BTN_DIGIT_9, _et11
    change  'w', 'x'
    change  'x', 'y'
    change  'y', 'z'
    change  'z', '9'
    change  'W', 'X'
    change  'X', 'Y'
    change  'Y', 'Z'
    change  'Z', '9'
    change  '9', '\x18'                 ; big Xi
    copyl   'W', INDF
    goto    EditDone

_et11
    testbut BTN_DIGIT_0, _et12
    change  '.', ','
    change  ',', '?'
    change  '?', '!'
    change  '!', '0'
    change  '0', '-'
    copyl   '.', INDF
    goto    EditDone

_et12
    testbut BTN_ARITY, EditDone         ; Swap case
    movfw   INDF
    sublw   'A'-1                       ; < 'A'?
    bc      EditDone
    movfw   INDF
  if GERMAN
    sublw   ']'                         ; < ''?
  else
    sublw   'Z'                         ; < 'Z'?
  endif
    bnc     _et13
    bsf     INDF, CASE_BIT              ; make lower case
    goto    EditDone
_et13
    movfw   INDF
    sublw   'a'-1                       ; < 'a'?
    bc      EditDone
    movfw   INDF
  if GERMAN
    sublw   '}'                         ; < ''?
  else
    sublw   'z'                         ; < 'z'?
  endif
    bnc     EditDone
    bcf     INDF, CASE_BIT              ; make upper case

EditDone
    copyl   0xff, rc_cmd
    return

;*************************************************************************************************
;*************************************************************************************************
;*
;* Map IR key codes to ASCII                
;*
;*************************************************************************************************
;*************************************************************************************************
    org 0xd00
    #include keymap.inc     

;*************************************************************************************************
;*************************************************************************************************
;*
;* Calculate moon phase (Yuck, rather code in Python :-)                    
;*
;*************************************************************************************************
;*************************************************************************************************
MoonPhase
    ;-----------------------------------
    ; Fetch remaining parms from RTC
    ;-----------------------------------
    seglower
    movlw   RTC_YEAR
    call    ReadRtc
    call    BCD2Bin    
    movwf   curr_year
    movlw   RTC_MOON
    call    ReadRtc
    movwf   moon_age
    movlw   RTC_HOUR
    call    ReadRtc
    segupper
    sublw   0x11                        ; Afternoon? (remember hour is BCD, so test is > 0x11)
    skpc
    incf    moon_age, F
    copy    curr_year, temp
    clrc    
    rrf     temp, F
    clrc
    rrf     temp, F                     ; temp := year/4
    movf    temp, F                     ; ==0 ?
    bz      _y4end
_y4loop
    movlw   .31                         ; (365*3+366)*2 % 59
    addwf   moon_age, F                 ; r += 31
    btfss   moon_age, 7                 ; normalize when too large
    goto    _y4next
    movlw   -.59 
    addwf   moon_age, F                 ; r -= 59
_y4next
    decfsz  temp, F
    goto    _y4loop
_y4end
    movlw   0x03
    andwf   curr_year, F                ; year %= 4
    bnz     _noLYCorr
    movlw   .3
    subwf   curr_month, W
    bc      _noLYCorr
    movlw   -.2
    addwf   moon_age, F                 ; r -= 2        
_noLYCorr
    movf    curr_year, F                ; year==0 ?
    bz      _yend
_yloop
    movlw   .22                         ; 365*2 % 59
    addwf   moon_age, F                 ; r += 22
    decfsz  curr_year, F
    goto    _yloop
_yend
    copyl   HIGH _monthMod59, PCLATH
    decf    curr_month, W
    call    _monthMod59
    addwf   moon_age, F                 ; r += monthMod59[month-1]
    btfss   moon_age, 7                 ; normalize when too large
    goto    _daynext
    movlw   -.59 
    addwf   moon_age, F                 ; r -= 59
_daynext
    movfw   curr_date
    addwf   moon_age, F
    addwf   moon_age, F                 ; r += 2*date
_normloop                               ; final normalization:
    movlw   .59                         ; r %= 59
    subwf   moon_age, W
    bnc     _normdone
    movwf   moon_age
    goto    _normloop
_normdone
    ;-----------------------------------
    ; Now select image to display
    ;-----------------------------------
    drawMoon .6,  ' ',  ' '             ;  6 New moon
    drawMoon .16, ' ',  ')'             ; 10 Waxing crescent
    drawMoon .19, ' ',  0x17            ;  3 1st quarter
    drawMoon .29, 0x0a, ')'             ; 10 Waxing gibbous
    drawMoon .36, '(',  ')'             ;  7 Full moon  
    drawMoon .46, '(', 0x0e             ; 10 Waning gibbous
    drawMoon .49, 0x16, ' '             ;  3 3rd quarter
    drawMoon .59, '(',  ' '             ; 10 Waning crescent
_moonDone
    return

_monthMod59:
    addwf   PCL, F
    retlw   .0                          ; half-days before Jan, 1 modulo 59
    retlw   .3                          ; half-days before Feb, 1 modulo 59
    retlw   .0
    retlw   .3
    retlw   .4
    retlw   .7
    retlw   .8
    retlw   .11
    retlw   .14
    retlw   .15 
    retlw   .18
    retlw   .19                         ; half-days before Dec, 1 modulo 59

;*************************************************************************************************
;*
;* Convert 16 bit number (num_hi, num_lo) to decimal in display buffer
;*
;*************************************************************************************************
ConvertDecimal
    copyl   0x27, num2                  ; num2 = 10000 (0x2710)
    copyl   0x10, num2+1
    call    Cmp16
    bc      _overflow
    movlw   '0'                         ; Faster than loading "0000" from EEPROM
    movwf   buffer
    movwf   buffer+1
    movwf   buffer+2
    movwf   buffer+3
    copyl   0x03, num2                  ; num2 = 1000 (0x03e8)
    copyl   0xe8, num2+1
_thousandsLoop
    call    Cmp16
    bnc     _hundreds
    call    Sub16
    incf    buffer, F
    goto    _thousandsLoop
_hundreds
    clrf    num2                        ; num2 = 100 (0x64)
    copyl   .100, num2+1
_hundredsLoop
    call    Cmp16
    bnc     _tens
    call    Sub16
    incf    buffer+1, F
    goto    _hundredsLoop
_tens
    copyl   .10, num2+1                 ; num2 = 10 (0x0a)
_tensLoop
    call    Cmp16
    bnc     _ones
    call    Sub16
    incf    buffer+2, F
    goto    _tensLoop
_ones
    movfw   num+1
    addwf   buffer+3, F
    return

_overflow
    movlw   .11                         ; "OVER"
    seglower
    goto    LoadWord

;----------------------------------------------------------
; Subtract 16 bit num2 from num
;----------------------------------------------------------
Sub16
    movfw   num2+1
    subwf   num+1, F
    skpc
    decf    num, F
    movfw   num2
    subwf   num, F
    return

;----------------------------------------------------------
; Test if num >= num2: C=1
;----------------------------------------------------------
Cmp16
    movfw   num2
    subwf   num, W
    skpz
    return
    movfw   num2+1
    subwf   num+1, W
    return

;*************************************************************************************************
;*
;* Convert 8 bit number in W to hex in display buffer (right 2 tubes)
;*
;*************************************************************************************************
ConvertHex
    movwf   temp
    andlw   0x0f                        ; low nibble
    movwf   temp1
    sublw   .9
    movfw   temp1
    skpc
    addlw   'A'-'0'-.10
    addlw   '0'
    movwf   buffer+3
    swapf   temp, W                     ; high nibble
    andlw   0x0f
    movwf   temp1
    sublw   .9
    movfw   temp1
    skpc
    addlw   'A'-'0'-.10
    addlw   '0'
    movwf   buffer+2
    return

;*************************************************************************************************
;*
;* Convert BCD seconds in W to cursor bit pattern:
;* We want a 'pendulum' effect, so we have to compute seconds modulo 6 
;*
;*************************************************************************************************
MakeSecCursor:
    movwf   temp
    copyl   HIGH _tenSec, PCLATH        ; Assume both jump tables in same page
    call    _tenSec                     ; Compute offset into seconds pattern table
    movwf   temp1
    movlw   0x0f
    andwf   temp, W                     ; Lower nibble of seconds
    addwf   temp1, F                    ; Add offset from upper nibble
    clrw
    btfsc   temp1, 3                    ; >=8 ?
    movlw   -.6                         ; To stay in table bounds
    addwf   temp1, W
    goto    _secPattern                 ; tail call

_tenSec:
    swapf   temp, W
    andlw   0x0f                        ; Extract 10 seconds to lower nibble
    addwf   PCL, F
    retlw   0
    retlw   4
    retlw   2
    retlw   0
    retlw   4
    retlw   2

_secPattern:
    addwf   PCL, F
    retlw   b'10000000'
    retlw   b'01000000'
    retlw   b'00100000'
    retlw   b'00010000'
    retlw   b'00100000'
    retlw   b'01000000'
    retlw   b'10000000'
    retlw   b'01000000'

;*************************************************************************************************
;*************************************************************************************************
;*
;* Convert ASCII code to segment patterns
;* code has been generated by Python module b7971.py
;*
;*************************************************************************************************
;*************************************************************************************************
    org 0xe7f
convert_lo
    addwf   PCL, F
  if GERMAN
    #include segments_uml_lo.inc
  else
    #include segments_lo.inc
  endif

    org 0xf7f
convert_hi
    addwf   PCL, F
  if GERMAN
    #include segments_uml_hi.inc
  else
    #include segments_hi.inc
  endif

;*************************************************************************************************
;*************************************************************************************************
;*
;* Internal EEPROM data tables               
;*
;*************************************************************************************************
;*************************************************************************************************
    org H'2100'
    ;-----------------------------------
    ; Some fixed texts
    ;-----------------------------------
    de "ROM?"                           ; 0     EEPROM invalid
  if GERMAN
    de "MON "                           ; 1     Montag
    de "DIE "                           ; 2     Dienstag
    de "MIT "                           ; 3     Mittwoch
    de "DON "                           ; 4     Donnerstag
    de "FRE "                           ; 5     Freitag
    de "SAM "                           ; 6     Samstag
    de "SON "                           ; 7     Sonntag
  else
    de "MON "                           ; 1     Monday
    de "TUE "                           ; 2     Tuesday
    de "WED "                           ; 3     Wednesday
    de "THU "                           ; 4     Thursday
    de "FRI "                           ; 5     Friday
    de "SAT "                           ; 6     Saturday
    de "SUN "                           ; 7     Sunday
  endif
    de "\x17UW\x16"                     ; 8     ready to upload EEPROM
    de "\x17SC\x16"                     ; 9     ready to upload clock
    de " OK "                           ; 10    Upload was OK
    de "OVER"                           ; 11    Overflow converting 16 bit int
                                        ;       to decimal
    de "****"                           ; 12    word chain broken
    de "MSG0"                           ; 13    default messages
    de "\x17UM\x16"                     ; 14    ready to upload messages
  if GERMAN
    de "V15G"                           ; 15    Version string
  else
    de "V15E"                           ; 15    Version string
  endif
    de "RTC?"                           ; 16    RTC not set
    de "F4LW"                           ; 17    Welcome message
    de "SP  "                           ; 18    Speed
    de "BR  "                           ; 19    Brightness
    de "RC  "                           ; 20    Remote control
    de "SER?"                           ; 21    No input on serial 
    de "MP  "                           ; 22    Moonphase
    de "----"                           ; 23    No sunrise/set at this date
    de "WRAP"                           ; 24    Wrap-around when switching DST or 
                                        ;       syncing clock 

rotatePat
    ;-----------------------------------
    ; Animation patterns for
    ; clockwise rotation
    ;-----------------------------------
    ;    LO mask      HI mask      LO sweep    HI sweep
    de b'00000000', b'00000000', b'00000001', b'00000000'
    de b'00000001', b'00000000', b'00000000', b'10000000'
    de b'01000001', b'10000000', b'00000000', b'01000000'
    de b'01100001', b'11000000', b'00000000', b'00100000'
    de b'01100001', b'11100000', b'00000000', b'00010000'
    de b'01110001', b'11110000', b'00000000', b'00001000'
    de b'01111001', b'11111000', b'00000000', b'00000100'
    de b'01111101', b'11111100', b'00000010', b'00000000'
    de b'11111111', b'11111100', b'00000001', b'00000000'
    de b'11111111', b'11111111', b'00000000', b'00000000'

scrollPat
    ;-----------------------------------
    ; Animation patterns for
    ; scrolling down
    ;-----------------------------------
    ;    LO mask      HI mask      LO sweep    HI sweep
    de b'00000000', b'00000000', b'10000000', b'00000000'
    de b'10000000', b'00000000', b'01000101', b'00000000'
    de b'11000111', b'10000000', b'00000000', b'01000100'
    de b'11000111', b'11000100', b'00101000', b'00010000'
    de b'11101111', b'11111100', b'00010000', b'00000000'
    de b'11111111', b'11111111', b'00000000', b'00000000'

lastPat                                 ; Marks end of animation patterns
    end

